"""Tests for ConnectedDriveVehicle."""
import unittest
from unittest import mock
from test import load_response_json, BackendMock, TEST_USERNAME, TEST_PASSWORD, TEST_REGION, \
    G31_VIN, F48_VIN, I01_VIN, I01_NOREX_VIN, F15_VIN, F45_VIN, F31_VIN, TEST_VEHICLE_DATA, \
    ATTRIBUTE_MAPPING, MISSING_ATTRIBUTES, ADDITIONAL_ATTRIBUTES, G30_PHEV_OS7_VIN, AVAILABLE_STATES_MAPPING

from bimmer_connected.vehicle import ConnectedDriveVehicle, DriveTrainType
from bimmer_connected.account import ConnectedDriveAccount


_VEHICLES = load_response_json('vehicles.json')['vehicles']
G31_VEHICLE = _VEHICLES[0]


class TestVehicle(unittest.TestCase):
    """Tests for ConnectedDriveVehicle."""

    def test_drive_train(self):
        """Tests around drive_train attribute."""
        vehicle = ConnectedDriveVehicle(None, G31_VEHICLE)
        self.assertEqual(DriveTrainType.CONVENTIONAL, vehicle.drive_train)

    def test_parsing_attributes(self):
        """Test parsing different attributes of the vehicle."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

        for vehicle in account.vehicles:
            print(vehicle.name)
            self.assertIsNotNone(vehicle.drive_train)
            self.assertIsNotNone(vehicle.name)
            self.assertIsNotNone(vehicle.has_internal_combustion_engine)
            self.assertIsNotNone(vehicle.has_hv_battery)
            self.assertIsNotNone(vehicle.drive_train_attributes)
            self.assertIsNotNone(vehicle.has_statistics_service)
            self.assertIsNotNone(vehicle.has_weekly_planner_service)
            self.assertIsNotNone(vehicle.has_destination_service)
            self.assertIsNotNone(vehicle.has_rangemap_service)

    def test_drive_train_attributes(self):
        """Test parsing different attributes of the vehicle."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

        for vehicle in account.vehicles:
            self.assertEqual(vehicle.vin in [G31_VIN, F48_VIN, F15_VIN, F45_VIN, F31_VIN, G30_PHEV_OS7_VIN],
                             vehicle.has_internal_combustion_engine)
            self.assertEqual(vehicle.vin in [I01_VIN, I01_NOREX_VIN, G30_PHEV_OS7_VIN],
                             vehicle.has_hv_battery)
            self.assertEqual(vehicle.vin in [I01_VIN],
                             vehicle.has_range_extender)

    def test_parsing_of_lsc_type(self):
        """Test parsing the lsc type field."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

        for vehicle in account.vehicles:
            self.assertIsNotNone(vehicle.lsc_type)

    def test_available_attributes(self):
        """Check that available_attributes returns exactly the arguments we have in our test data."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

        for vin, dirname in TEST_VEHICLE_DATA.items():
            vehicle = account.get_vehicle(vin)
            print(vehicle.name)
            status_data = load_response_json('{}/status.json'.format(dirname))
            existing_attributes = status_data['vehicleStatus'].keys()
            existing_attributes = sorted([ATTRIBUTE_MAPPING.get(a, a) for a in existing_attributes
                                          if a not in MISSING_ATTRIBUTES])
            expected_attributes = sorted([a for a in vehicle.available_attributes if a not in ADDITIONAL_ATTRIBUTES])
            self.assertListEqual(existing_attributes, expected_attributes)

    def test_available_state_services(self):
        """Check that available_attributes returns exactly the arguments we have in our test data."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

        vehicles = load_response_json('vehicles.json')

        for test_vehicle in vehicles['vehicles']:
            vehicle = account.get_vehicle(test_vehicle['vin'])
            print(vehicle.name)

            services_to_check = {
                k: v
                for k, v in test_vehicle.items()
                if k in list(AVAILABLE_STATES_MAPPING)
            }

            available_services = ['STATUS']
            for key, value in services_to_check.items():
                if AVAILABLE_STATES_MAPPING[key].get(value):
                    available_services += AVAILABLE_STATES_MAPPING[key][value]
            if vehicle.drive_train != DriveTrainType.CONVENTIONAL:
                available_services += ['EFFICIENCY', 'NAVIGATION']

            self.assertListEqual(sorted(vehicle.available_state_services), sorted(available_services))



"""Models the state of a vehicle."""

import datetime
import logging
from enum import Enum
from typing import List

from bimmer_connected.const import SERVICE_STATUS

_LOGGER = logging.getLogger(__name__)


LIDS = ['doorDriverFront', 'doorPassengerFront', 'doorDriverRear', 'doorPassengerRear',
        'hood', 'trunk']

WINDOWS = ['windowDriverFront', 'windowPassengerFront', 'windowDriverRear', 'windowPassengerRear', 'rearWindow',
           'sunroof']


class LidState(Enum):
    """Possible states of the hatch, trunk, doors, windows, sun roof."""
    CLOSED = 'CLOSED'
    OPEN = 'OPEN'
    OPEN_TILT = 'OPEN_TILT'
    INTERMEDIATE = 'INTERMEDIATE'
    INVALID = 'INVALID'


class LockState(Enum):
    """Possible states of the door locks."""
    LOCKED = 'LOCKED'
    SECURED = 'SECURED'
    SELECTIVE_LOCKED = 'SELECTIVE_LOCKED'
    UNLOCKED = 'UNLOCKED'


class ParkingLightState(Enum):
    """Possible states of the parking lights"""
    LEFT = 'LEFT'
    RIGHT = 'RIGHT'
    OFF = 'OFF'


class ConditionBasedServiceStatus(Enum):
    """Status of the condition based services."""
    OK = 'OK'
    OVERDUE = 'OVERDUE'
    PENDING = 'PENDING'


class ChargingState(Enum):
    """Charging state of electric vehicle."""
    CHARGING = 'CHARGING'
    ERROR = 'ERROR'
    FINISHED_FULLY_CHARGED = 'FINISHED_FULLY_CHARGED'
    FINISHED_NOT_FULL = 'FINISHED_NOT_FULL'
    INVALID = 'INVALID'
    NOT_CHARGING = 'NOT_CHARGING'
    WAITING_FOR_CHARGING = 'WAITING_FOR_CHARGING'


class CheckControlMessage:
    """Check control message sent from the server.

    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, ccm_dict: dict):
        self._ccm_dict = ccm_dict

    @property
    def description_long(self) -> str:
        """Long description of the check control message."""
        return self._ccm_dict["ccmDescriptionLong"]

    @property
    def description_short(self) -> str:
        """Short description of the check control message."""
        return self._ccm_dict["ccmDescriptionShort"]

    @property
    def ccm_id(self) -> int:
        """id of the check control message."""
        return int(self._ccm_dict["ccmId"])

    @property
    def mileage(self) -> int:
        """Mileage of the vehicle when the check control message appeared."""
        return int(self._ccm_dict["ccmMileage"])


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'VehicleStatus', *args, **kwargs):
        # pylint: disable=protected-access
        if self._state.attributes[SERVICE_STATUS] is None:
            raise ValueError('No data available for vehicle status!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class VehicleStatus:  # pylint: disable=too-many-public-methods
    """Models the status of a vehicle."""

    def __init__(self, state):
        """Constructor."""
        self._state = state

    @property
    @backend_parameter
    def attributes(self) -> dict:
        """Retrieve all attributes from the sever.

        This does not parse the results in any way.
        """
        return self._state.attributes[SERVICE_STATUS]

    @property
    @backend_parameter
    def timestamp(self) -> datetime.datetime:
        """Get the timestamp when the data was recorded."""
        return self._parse_datetime(self._state.attributes[SERVICE_STATUS]['updateTime'])

    @property
    @backend_parameter
    def gps_position(self) -> (float, float):
        """Get the last known position of the vehicle.

        Returns a tuple of (latitude, longitude).
        This only provides data, if the vehicle tracking is enabled!
        """
        pos = self._state.attributes[SERVICE_STATUS]['position']
        if not self.is_vehicle_tracking_enabled:
            _LOGGER.warning('Positioning status is %s', pos['status'])
            return None
        pos = self._state.attributes[SERVICE_STATUS]['position']
        return float(pos['lat']), float(pos['lon'])

    @property
    @backend_parameter
    def is_vehicle_tracking_enabled(self) -> bool:
        """Check if the position tracking of the vehicle is enabled.

        The server return "OK" if tracking is enabled and "DRIVER_DISABLED" if it is disabled in the vehicle.
        """
        return self._state.attributes[SERVICE_STATUS]['position']['status'] == 'OK'

    @property
    @backend_parameter
    def mileage(self) -> int:
        """Get the mileage of the vehicle.

        Returns a tuple of (value, unit_of_measurement)
        """
        return int(self._state.attributes[SERVICE_STATUS]['mileage'])

    @property
    @backend_parameter
    def remaining_range_fuel(self) -> int:
        """Get the remaining range of the vehicle on fuel.

        Returns a tuple of (value, unit_of_measurement)
        """
        return int(self._state.attributes[SERVICE_STATUS]['remainingRangeFuel'])

    @property
    @backend_parameter
    def remaining_fuel(self) -> int:
        """Get the remaining fuel of the vehicle.

        Returns a tuple of (value, unit_of_measurement)
        """
        return int(self._state.attributes[SERVICE_STATUS]['remainingFuel'])

    @property
    @backend_parameter
    def lids(self) -> List['Lid']:
        """Get all lids (doors+hatch+trunk) of the car."""
        result = []
        for lid in LIDS:
            if lid in self._state.attributes[SERVICE_STATUS] and \
                    self._state.attributes[SERVICE_STATUS][lid] != LidState.INVALID.value:
                result.append(Lid(self, lid))
        return result

    @property
    def open_lids(self) -> List['Lid']:
        """Get all open lids of the car."""
        return [lid for lid in self.lids if not lid.is_closed]

    @property
    def all_lids_closed(self) -> bool:
        """Check if all lids are closed."""
        return len(list(self.open_lids)) == 0

    @property
    @backend_parameter
    def windows(self) -> List['Window']:
        """Get all windows (doors+sun roof) of the car."""
        result = []
        for window in WINDOWS:
            if window in self._state.attributes[SERVICE_STATUS] and \
                    self._state.attributes[SERVICE_STATUS][window] != LidState.INVALID.value:
                result.append(Window(self, window))
        return result

    @property
    def open_windows(self) -> List['Window']:
        """Get all open windows of the car."""
        return [lid for lid in self.windows if not lid.is_closed]

    @property
    def all_windows_closed(self) -> bool:
        """Check if all windows are closed."""
        return len(list(self.open_windows)) == 0

    @property
    @backend_parameter
    def door_lock_state(self) -> LockState:
        """Get state of the door locks."""
        return LockState(self._state.attributes[SERVICE_STATUS]['doorLockState'])

    @property
    @backend_parameter
    def last_update_reason(self) -> str:
        """The reason for the last state update"""
        return self._state.attributes[SERVICE_STATUS]['updateReason']

    @property
    @backend_parameter
    def last_charging_end_result(self) -> str:
        """Get the last charging end result"""
        return self._state.attributes[SERVICE_STATUS]['lastChargingEndResult']

    @property
    @backend_parameter
    def connection_status(self) -> str:
        """Get status of the connection"""
        return self._state.attributes[SERVICE_STATUS]['connectionStatus']

    @property
    @backend_parameter
    def condition_based_services(self) -> List['ConditionBasedServiceReport']:
        """Get status of the condition based services."""
        return [ConditionBasedServiceReport(s) for s in self._state.attributes[SERVICE_STATUS]['cbsData']]

    @property
    def are_all_cbs_ok(self) -> bool:
        """Check if the status of all condition based services is "OK"."""
        for cbs in self.condition_based_services:
            if cbs.state != ConditionBasedServiceStatus.OK:
                return False
        return True

    @property
    @backend_parameter
    def parking_lights(self) -> ParkingLightState:
        """Get status of parking lights.

        :returns None if status is unknown.
        """
        return ParkingLightState(self._state.attributes[SERVICE_STATUS]['parkingLight'])

    @property
    def are_parking_lights_on(self) -> bool:
        """Get status of parking lights.

        :returns None if status is unknown.
        """
        lights = self.parking_lights
        if lights is None:
            return None
        return lights != ParkingLightState.OFF

    @staticmethod
    def _parse_datetime(date_str: str) -> datetime.datetime:
        """Convert a time string into datetime."""
        date_format = "%Y-%m-%dT%H:%M:%S%z"
        return datetime.datetime.strptime(date_str, date_format)

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self._state.attributes[SERVICE_STATUS][item]

    @property
    @backend_parameter
    def remaining_range_electric(self) -> int:
        """Remaining range on battery, in kilometers."""
        return int(self._state.attributes[SERVICE_STATUS]['remainingRangeElectric'])

    @property
    @backend_parameter
    def remaining_range_total(self) -> int:
        """Get the total remaining range of the vehicle in kilometers.

        That is electrical range + fuel range.
        """
        result = 0
        if self.remaining_range_electric is not None:
            result += self.remaining_range_electric
        if self.remaining_range_fuel is not None:
            result += self.remaining_range_fuel
        return result

    @property
    @backend_parameter
    def max_range_electric(self) -> int:
        """ This can change with driving style and temperature in kilometers."""
        return int(self._state.attributes[SERVICE_STATUS]['maxRangeElectric'])

    @property
    @backend_parameter
    def charging_status(self) -> ChargingState:
        """Charging state of the vehicle."""
        state = self._state.attributes[SERVICE_STATUS]['chargingStatus']
        return ChargingState(state)

    @property
    @backend_parameter
    def charging_time_remaining(self) -> datetime.timedelta:
        """Get the remaining charging time."""
        minutes = self._state.attributes[SERVICE_STATUS]['chargingTimeRemaining']
        return datetime.timedelta(minutes=minutes)

    @property
    @backend_parameter
    def charging_level_hv(self) -> int:
        """State of charge of the high voltage battery in percent."""
        return int(self._state.attributes[SERVICE_STATUS]['chargingLevelHv'])

    @property
    @backend_parameter
    def fuel_percent(self) -> int:
        """State of fuel in percent."""
        return int(self._state.attributes[SERVICE_STATUS]['fuelPercent'])

    @property
    @backend_parameter
    def check_control_messages(self) -> List[CheckControlMessage]:
        """List of check control messages."""
        # TO DO change this in HA binary_sensor.py first
        # messages = self._state.attributes[SERVICE_STATUS]['STATUS'].get('checkControlMessages', [])
        # return [CheckControlMessage(m) for m in messages]
        return self._state.attributes[SERVICE_STATUS].get('checkControlMessages', [])

    @property
    @backend_parameter
    def has_check_control_messages(self) -> bool:
        """Return true if any check control message is present."""
        return len(self.check_control_messages) > 0


class Lid:  # pylint: disable=too-few-public-methods
    """A lid of the vehicle.

    Lids are: Doors + Trunk + Hatch
    """

    def __init__(self, vehicle_status: VehicleStatus, name: str):
        #: name of the lid
        self.name = name
        self._vehicle_status = vehicle_status

    @property
    def state(self):
        """Get the current state of the lid."""
        return LidState(getattr(self._vehicle_status, self.name))

    @property
    def is_closed(self) -> bool:
        """Check if the lid is closed."""
        return self.state == LidState.CLOSED

    def __str__(self) -> str:
        return '{}: {}'.format(self.name, self._vehicle_status)


class Window(Lid):  # pylint: disable=too-few-public-methods
    """A window of the vehicle.

    A window can be a normal window of the car or the sun roof.
    """


class ConditionBasedServiceReport:  # pylint: disable=too-few-public-methods
    """Entry in the list of condition based services."""

    def __init__(self, data: dict):

        #: date when the service is due
        self.due_date = self._parse_date(data.get('cbsDueDate'))

        #: status of the service
        self.state = ConditionBasedServiceStatus(data['cbsState'])

        #: service type
        self.service_type = data['cbsType']

        #: distance when the service is due
        self.due_distance = None
        if 'cbsRemainingMileage' in data:
            self.due_distance = int(data['cbsRemainingMileage'])

        #: description of the required service
        self.description = data['cbsDescription']

    @staticmethod
    def _parse_date(datestr: str) -> datetime.datetime:
        if datestr is None:
            return None
        formats = [
            '%Y-%m',
            '%m.%Y',
        ]
        for date_format in formats:
            try:
                date = datetime.datetime.strptime(datestr, date_format)
                return date.replace(day=1)
            except ValueError:
                pass
        _LOGGER.error('Unknown time format for CBS: %s', datestr)
        return None



"""Test for LastDestinations"""
import unittest
from unittest import mock
from test import load_response_json
from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.state import VehicleState
from bimmer_connected.const import SERVICE_DESTINATIONS
from bimmer_connected.last_destinations import DestinationType

I01_TEST_DATA = load_response_json('I01_NOREX/last_destinations.json')


class TestState(unittest.TestCase):
    """Test for LastDestinations."""

    # pylint: disable=protected-access

    def test_parse_i01(self):
        """Test if the parsing of the attributes is working."""
        account = mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_DESTINATIONS] = I01_TEST_DATA['destinations']
        self.assertEqual(DestinationType.DESTINATION, state.last_destinations.last_destinations[0].destination_type)
        self.assertEqual(51.53053283691406, state.last_destinations.last_destinations[0].latitude)
        self.assertEqual(-0.08362331241369247, state.last_destinations.last_destinations[0].longitude)
        self.assertEqual('UNITED KINGDOM', state.last_destinations.last_destinations[0].country)
        self.assertEqual('LONDON', state.last_destinations.last_destinations[0].city)
        self.assertEqual('PITFIELD STREET', state.last_destinations.last_destinations[0].street)
        self.assertEqual('2015-09-25T08:06:11+0200', state.last_destinations.last_destinations[0].created_at)



"""Test for Efficiency"""
import unittest
from unittest import mock
from test import load_response_json
from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.state import VehicleState
from bimmer_connected.const import SERVICE_EFFICIENCY

G30_PHEV_OS7_TEST_DATA = load_response_json('G30_PHEV_OS7/efficiency.json')


class TestState(unittest.TestCase):
    """Test for Efficiency."""

    # pylint: disable=protected-access

    def test_parse_g30_phev_os7(self):
        """Test if the parsing of the attributes is working."""
        account = mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_EFFICIENCY] = G30_PHEV_OS7_TEST_DATA
        self.assertEqual('PHEV', state.efficiency.model_type)
        self.assertEqual(0, state.efficiency.efficiency_quotient)
        self.assertEqual('LASTTRIP_DELTA_KM', state.efficiency.last_trip_list[0].name)
        self.assertEqual('KM', state.efficiency.last_trip_list[0].unit)
        self.assertEqual('--', state.efficiency.last_trip_list[0].last_trip)
        self.assertEqual('TIMESTAMP_STATISTICS_RESET', state.efficiency.life_time_list[2].name)
        self.assertIsNone(state.efficiency.life_time_list[2].unit)
        self.assertEqual('12.01.2020', state.efficiency.life_time_list[2].life_time)
        self.assertEqual('DRIVING_MODE', state.efficiency.characteristic_list[1].characteristic)
        self.assertEqual(0, state.efficiency.characteristic_list[1].quantity)



"""Test for RangeMaps"""
import unittest
from unittest import mock
from test import load_response_json
from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.state import VehicleState
from bimmer_connected.const import SERVICE_RANGEMAP
from bimmer_connected.range_maps import RangeMapType, RangeMapQuality

I01_TEST_DATA = load_response_json('I01_NOREX/range_maps.json')


class TestState(unittest.TestCase):
    """Test for LastDestinations."""

    # pylint: disable=protected-access

    def test_parse_i01(self):
        """Test if the parsing of the attributes is working."""
        account = mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_RANGEMAP] = I01_TEST_DATA['rangemap']
        self.assertEqual(RangeMapQuality.AVERAGE, state.range_maps.range_map_quality)
        self.assertEqual(51.123456, state.range_maps.range_map_center.latitude)
        self.assertEqual(-1.2345678, state.range_maps.range_map_center.longitude)
        self.assertEqual(RangeMapType.ECO_PRO_PLUS, state.range_maps.range_maps[0].range_map_type)
        self.assertEqual(51.6991281509399, state.range_maps.range_maps[0].polyline[0].latitude)
        self.assertEqual(-2.00423240661621, state.range_maps.range_maps[0].polyline[0].longitude)



"""Test for Navigation"""
import unittest
from unittest import mock
from test import load_response_json
from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.state import VehicleState
from bimmer_connected.const import SERVICE_NAVIGATION

G30_PHEV_OS7_TEST_DATA = load_response_json('G30_PHEV_OS7/navigation.json')


class TestState(unittest.TestCase):
    """Test for Navigation."""

    # pylint: disable=protected-access

    def test_parse_g30_phev_os7(self):
        """Test if the parsing of the attributes is working."""
        account = mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_NAVIGATION] = G30_PHEV_OS7_TEST_DATA
        self.assertEqual(52.82273, state.navigation.latitude)
        self.assertEqual(8.8276, state.navigation.longitude)
        self.assertEqual('DEU', state.navigation.iso_country_code)
        self.assertEqual(1.4, state.navigation.aux_power_regular)
        self.assertEqual(1.2, state.navigation.aux_power_eco_pro)
        self.assertEqual(0.4, state.navigation.aux_power_eco_pro_plus)
        self.assertEqual(4.98700008381234, state.navigation.soc)
        self.assertEqual(9.48, state.navigation.soc_max)
        self.assertIsNone(state.navigation.eco)
        self.assertIsNone(state.navigation.norm)
        self.assertIsNone(state.navigation.eco_ev)
        self.assertIsNone(state.navigation.norm_ev)
        self.assertIsNone(state.navigation.vehicle_mass)
        self.assertIsNone(state.navigation.k_acc_reg)
        self.assertIsNone(state.navigation.k_dec_reg)
        self.assertIsNone(state.navigation.k_acc_eco)
        self.assertIsNone(state.navigation.k_dec_eco)
        self.assertIsNone(state.navigation.k_up)
        self.assertIsNone(state.navigation.k_down)
        self.assertIsNone(state.navigation.drive_train)
        self.assertFalse(state.navigation.pending_update)
        self.assertFalse(state.navigation.vehicle_tracking)



"""Test for AllTrips."""
import unittest
from unittest import mock
from test import load_response_json
from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.state import VehicleState
from bimmer_connected.const import SERVICE_ALL_TRIPS

I01_TEST_DATA = load_response_json('I01_NOREX/all_trips.json')


class TestState(unittest.TestCase):
    """Test for VehicleState."""

    # pylint: disable=protected-access

    def test_parse_i01(self):
        """Test if the parsing of the attributes is working."""
        account = mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_ALL_TRIPS] = I01_TEST_DATA['allTrips']
        self.assertEqual('1970-01-01T01:00:00+0100', state.all_trips.reset_date)
        self.assertEqual(35820, state.all_trips.battery_size_max)
        self.assertEqual(87.58, state.all_trips.saved_co2)
        self.assertEqual(515.177, state.all_trips.saved_co2_green_energy)
        self.assertEqual(0, state.all_trips.total_saved_fuel)

        self.assertEqual(0, state.all_trips.average_electric_consumption.community_low)
        self.assertEqual(16.33, state.all_trips.average_electric_consumption.community_average)
        self.assertEqual(35.53, state.all_trips.average_electric_consumption.community_high)
        self.assertEqual(14.76, state.all_trips.average_electric_consumption.user_average)

        self.assertEqual(0, state.all_trips.average_recopuration.community_low)
        self.assertEqual(3.76, state.all_trips.average_recopuration.community_average)
        self.assertEqual(14.03, state.all_trips.average_recopuration.community_high)
        self.assertEqual(2.3, state.all_trips.average_recopuration.user_average)

        self.assertEqual(121.58, state.all_trips.chargecycle_range.community_average)
        self.assertEqual(200, state.all_trips.chargecycle_range.community_high)
        self.assertEqual(72.62, state.all_trips.chargecycle_range.user_average)
        self.assertEqual(135, state.all_trips.chargecycle_range.user_high)
        self.assertEqual(60, state.all_trips.chargecycle_range.user_current_charge_cycle)

        self.assertEqual(1, state.all_trips.total_electric_distance.community_low)
        self.assertEqual(12293.65, state.all_trips.total_electric_distance.community_average)
        self.assertEqual(77533.6, state.all_trips.total_electric_distance.community_high)
        self.assertEqual(3158.66, state.all_trips.total_electric_distance.user_total)

        self.assertEqual(0, state.all_trips.average_combined_consumption.community_low)
        self.assertEqual(1.21, state.all_trips.average_combined_consumption.community_average)
        self.assertEqual(6.2, state.all_trips.average_combined_consumption.community_high)
        self.assertEqual(0.36, state.all_trips.average_combined_consumption.user_average)



"""Models the range maps of a vehicle."""

import logging
from typing import List
from enum import Enum

from bimmer_connected.const import SERVICE_RANGEMAP

_LOGGER = logging.getLogger(__name__)


class RangeMapServices(Enum):
    """Range map services."""
    POLYGON = 'RANGE_POLYGON'
    CIRCLE = 'RANGE_CIRCLE'


class RangeMapType(Enum):
    """Range map types."""
    ECO_PRO_PLUS = 'ECO_PRO_PLUS'
    COMFORT = 'COMFORT'


class RangeMapQuality(Enum):
    """Range map types."""
    AVERAGE = 'AVERAGE'


class MapPoint:
    """
    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, mp_dict: dict):
        self._mp_dict = mp_dict

    @property
    def latitude(self) -> float:
        """latitude of this point."""
        return float(self._mp_dict["lat"])

    @property
    def longitude(self) -> float:
        """longitude of this point."""
        return float(self._mp_dict["lon"])


class RangeMap:
    """
    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, range_dict: dict):
        self._range_dict = range_dict

    @property
    def range_map_type(self) -> RangeMapType:
        """Type of the range map."""
        return RangeMapType(self._range_dict["type"])

    @property
    def polyline(self) -> List[MapPoint]:
        """polylines of this range map."""
        pol_list = []
        for map_point in self._range_dict["polyline"]:
            pol_list.append(MapPoint(map_point))
        return pol_list


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'RangeMaps', *args, **kwargs):
        # pylint: disable=protected-access
        if self._state.attributes[SERVICE_RANGEMAP] is None:
            raise ValueError('No data available for range maps!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class RangeMaps:  # pylint: disable=too-many-public-methods
    """Models the range maps of a vehicle."""

    def __init__(self, state):
        """Constructor."""
        self._state = state

    @property
    @backend_parameter
    def attributes(self) -> dict:
        """Retrieve all attributes from the sever.

        This does not parse the results in any way.
        """
        return self._state.attributes[SERVICE_RANGEMAP]

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self._state.attributes[SERVICE_RANGEMAP][item]

    @property
    @backend_parameter
    def range_map_quality(self) -> RangeMapQuality:
        """Type of the range map."""
        return RangeMapQuality(self._state.attributes[SERVICE_RANGEMAP]["quality"])

    @property
    @backend_parameter
    def range_map_center(self) -> RangeMapQuality:
        """Type of the range map."""
        return MapPoint(self._state.attributes[SERVICE_RANGEMAP]["center"])

    @property
    @backend_parameter
    def range_maps(self) -> List[RangeMap]:
        """Get the list range maps."""
        range_maps_list = []
        for dest in self._state.attributes[SERVICE_RANGEMAP]["rangemaps"]:
            range_maps_list.append(RangeMap(dest))
        return range_maps_list



"""Test for LastTrip"""
import unittest
from unittest import mock
from test import load_response_json
from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.state import VehicleState
from bimmer_connected.const import SERVICE_LAST_TRIP

I01_TEST_DATA = load_response_json('I01_NOREX/last_trip.json')


class TestState(unittest.TestCase):
    """Test for LastTrip."""

    # pylint: disable=protected-access

    def test_parse_i01(self):
        """Test if the parsing of the attributes is working."""
        account = mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_LAST_TRIP] = I01_TEST_DATA['lastTrip']
        self.assertEqual(0.53, state.last_trip.efficiencyValue)
        self.assertEqual(141, state.last_trip.total_distance)
        self.assertEqual(100.1, state.last_trip.electric_distance)
        self.assertEqual(16.6, state.last_trip.average_electric_consumption)
        self.assertEqual(2, state.last_trip.average_recuperation)
        self.assertEqual(0, state.last_trip.driving_mode_value)
        self.assertEqual(0.39, state.last_trip.acceleration_value)
        self.assertEqual(0.81, state.last_trip.anticipation_value)
        self.assertEqual(0.79, state.last_trip.total_consumption_value)
        self.assertEqual(0.66, state.last_trip.auxiliary_consumption_value)
        self.assertEqual(1.9, state.last_trip.average_combined_consumption)
        self.assertEqual(71, state.last_trip.electric_distance_ratio)
        self.assertEqual(0, state.last_trip.saved_fuel)
        self.assertEqual('2015-12-01T20:44:00+0100', state.last_trip.date)
        self.assertEqual(124, state.last_trip.duration)



"""Test for remote_services."""
import unittest
from unittest import mock
import datetime
from test import BackendMock, TEST_USERNAME, TEST_PASSWORD, TEST_REGION, G31_VIN, load_response_json, \
    POI_DATA, POI_REQUEST, MESSAGE_DATA, MESSAGE_REQUEST

from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.remote_services import RemoteServiceStatus, ExecutionState, PointOfInterest, Message
from bimmer_connected import remote_services

_RESPONSE_UNKNOWN = 'G31_NBTevo/flash_unknown.json'
_RESPONSE_INITIATED = 'G31_NBTevo/flash_initiated.json'
_RESPONSE_PENDING = 'G31_NBTevo/flash_pending.json'
_RESPONSE_DELIVERED = 'G31_NBTevo/flash_delivered.json'
_RESPONSE_EXECUTED = 'G31_NBTevo/flash_executed.json'
_MSG_EXECUTED = 'G31_NBTevo/msg_executed.json'


class TestRemoteServices(unittest.TestCase):
    """Test for remote_services."""

    # pylint: disable=protected-access

    def test_parse_timestamp(self):
        """Test parsing the timestamp format."""
        timestamp = RemoteServiceStatus._parse_timestamp("2018-02-11T15:10:39.465+01")
        expected = datetime.datetime(year=2018, month=2, day=11, hour=15, minute=10, second=39, microsecond=465000)
        self.assertEqual(expected, timestamp)

    def test_states(self):
        """Test parsing the different response types."""
        rss = RemoteServiceStatus(load_response_json(_RESPONSE_UNKNOWN))
        self.assertEqual(ExecutionState.UNKNOWN, rss.state)

        rss = RemoteServiceStatus(load_response_json(_RESPONSE_INITIATED))
        self.assertEqual(ExecutionState.INITIATED, rss.state)

        rss = RemoteServiceStatus(load_response_json(_RESPONSE_PENDING))
        self.assertEqual(ExecutionState.PENDING, rss.state)

        rss = RemoteServiceStatus(load_response_json(_RESPONSE_DELIVERED))
        self.assertEqual(ExecutionState.DELIVERED, rss.state)

        rss = RemoteServiceStatus(load_response_json(_RESPONSE_EXECUTED))
        self.assertEqual(ExecutionState.EXECUTED, rss.state)

    def test_trigger_remote_services(self):
        """Test executing a remote light flash."""
        remote_services._POLLING_CYCLE = 0
        remote_services._UPDATE_AFTER_REMOTE_SERVICE_DELAY = 0

        services = [
            ('LIGHT_FLASH', 'trigger_remote_light_flash', False),
            ('DOOR_LOCK', 'trigger_remote_door_lock', True),
            ('DOOR_UNLOCK', 'trigger_remote_door_unlock', True),
            ('CLIMATE_NOW', 'trigger_remote_air_conditioning', True),
            ('VEHICLE_FINDER', 'trigger_remote_vehicle_finder', True),
            ('HORN_BLOW', 'trigger_remote_horn', False),
            ('SEND_MESSAGE', 'trigger_send_message', False),
            ('SEND_POI', 'trigger_send_poi', False),
        ]

        for service, call, triggers_update in services:
            backend_mock = BackendMock()
            backend_mock.setup_default_vehicles()

            backend_mock.add_response('https://.+/webapi/v1/user/vehicles/{vin}/executeService'.format(vin=G31_VIN),
                                      data_files=[_RESPONSE_INITIATED])

            backend_mock.add_response(
                r'https://.+/webapi/v1/user/vehicles/{vin}/serviceExecutionStatus\?serviceType={service_type}'.format(
                    vin=G31_VIN, service_type=service),
                r'https://.+/webapi/v1/user/vehicles/{vin}/status'.format(
                    vin=G31_VIN),
                data_files=[
                    _RESPONSE_UNKNOWN,
                    _RESPONSE_PENDING,
                    _RESPONSE_PENDING,
                    _RESPONSE_DELIVERED,
                    _RESPONSE_DELIVERED,
                    _RESPONSE_EXECUTED])

            backend_mock.add_response(
                r'https://.+/webapi/v1/user/vehicles/{vin}/status'.format(
                    vin=G31_VIN),
                data_files=[_RESPONSE_EXECUTED])

            backend_mock.add_response(
                r'https://.+/webapi/v1/user/vehicles/{vin}/sendpoi'.format(
                    vin=G31_VIN),
                data_files=[_MSG_EXECUTED],
                status_code=204)

            with mock.patch('bimmer_connected.account.requests', new=backend_mock):
                account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
                mock_listener = mock.Mock(return_value=None)
                account.add_update_listener(mock_listener)
                vehicle = account.get_vehicle(G31_VIN)

                if service == 'SEND_MESSAGE':
                    response = getattr(vehicle.remote_services, call)(MESSAGE_DATA)
                elif service == 'SEND_POI':
                    response = getattr(vehicle.remote_services, call)(POI_DATA)
                else:
                    response = getattr(vehicle.remote_services, call)()
                self.assertEqual(ExecutionState.EXECUTED, response.state)

                if triggers_update:
                    mock_listener.assert_called_once_with()
                else:
                    mock_listener.assert_not_called()

    def test_get_remote_service_status(self):
        """Test get_remove_service_status method."""
        backend_mock = BackendMock()

        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
            vehicle = account.get_vehicle(G31_VIN)
            with self.assertRaises(IOError):
                vehicle.remote_services._get_remote_service_status(remote_services._Services.REMOTE_LIGHT_FLASH)

            backend_mock.add_response(
                r'https://.+/webapi/v1/user/vehicles/{vin}/serviceExecutionStatus\?.+'.format(vin=G31_VIN),
                data_files=['G31_NBTevo/flash_executed.json'])
            status = vehicle.remote_services._get_remote_service_status(remote_services._Services.REMOTE_LIGHT_FLASH)
            self.assertEqual(ExecutionState.EXECUTED, status.state)

    def test_parsing_of_poi_min_attributes(self):
        """Check that a PointOfInterest can be constructed using only latitude & longitude."""
        poi = PointOfInterest(POI_DATA["lat"], POI_DATA["lon"])
        msg = Message.from_poi(poi)
        self.assertEqual(msg.as_server_request, POI_REQUEST["min"])

    def test_parsing_of_poi_all_attributes(self):
        """Check that a PointOfInterest can be constructed using all attributes."""
        poi = PointOfInterest(POI_DATA["lat"], POI_DATA["lon"], name=POI_DATA["name"],
                              additional_info=POI_DATA["additional_info"], street=POI_DATA["street"],
                              city=POI_DATA["city"], postal_code=POI_DATA["postal_code"],
                              country=POI_DATA["country"], website=POI_DATA["website"],
                              phone_numbers=POI_DATA["phone_numbers"])
        msg = Message.from_poi(poi)
        self.assertEqual(msg.as_server_request, POI_REQUEST["all"])

    def test_parsing_of_message_min_attributes(self):
        """Check that a Message can be constructed using text."""
        msg = Message.from_text(MESSAGE_DATA["text"])
        self.assertEqual(msg.as_server_request, MESSAGE_REQUEST["min"])

    def test_parsing_of_message_all_attributes(self):
        """Check that a Message can be constructed using text."""
        msg = Message.from_text(MESSAGE_DATA["text"], MESSAGE_DATA["subject"])
        self.assertEqual(msg.as_server_request, MESSAGE_REQUEST["all"])



"""Models the efficiency of a vehicle."""

import logging
from typing import List

from bimmer_connected.const import SERVICE_EFFICIENCY

_LOGGER = logging.getLogger(__name__)


class LastTrip:
    """Last Trip.

    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, trip_dict: dict):
        self._trip_dict = trip_dict

    @property
    def name(self) -> str:
        """Name."""
        return self._trip_dict["name"]

    @property
    def unit(self) -> str:
        """Unit."""
        return self._trip_dict["unit"]

    @property
    def last_trip(self) -> str:
        """Last Trip."""
        return self._trip_dict["lastTrip"]


def backend_parameter_life_time(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'LifeTime', *args, **kwargs):
        # pylint: disable=protected-access
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class LifeTime:
    """Life Time.

    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, time_dict: dict):
        self._time_dict = time_dict

    @property
    @backend_parameter_life_time
    def name(self) -> str:
        """Name."""
        return self._time_dict["name"]

    @property
    @backend_parameter_life_time
    def unit(self) -> str:
        """Unit."""
        return self._time_dict["unit"]

    @property
    @backend_parameter_life_time
    def life_time(self) -> str:
        """Life Time."""
        return self._time_dict["lifeTime"]


class Characteristic:
    """Characteristic.

    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, char_dict: dict):
        self._char_dict = char_dict

    @property
    def characteristic(self) -> str:
        """Characteristic."""
        return self._char_dict["characteristic"]

    @property
    def quantity(self) -> int:
        """Quantity."""
        return int(self._char_dict["quantity"])


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'Efficiency', *args, **kwargs):
        # pylint: disable=protected-access
        if self._state.attributes[SERVICE_EFFICIENCY] is None:
            raise ValueError('No data available for vehicles efficiency!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class Efficiency:  # pylint: disable=too-many-public-methods
    """Models the navigation of a vehicle."""

    def __init__(self, state):
        """Constructor."""
        self._state = state

    @property
    @backend_parameter
    def attributes(self) -> dict:
        """Retrieve all attributes from the sever.

        This does not parse the results in any way.
        """
        return self._state.attributes[SERVICE_EFFICIENCY]

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self._state.attributes[SERVICE_EFFICIENCY][item]

    @property
    @backend_parameter
    def model_type(self) -> str:
        """Returns the model type."""
        return self._state.attributes[SERVICE_EFFICIENCY]['modelType']

    @property
    @backend_parameter
    def efficiency_quotient(self) -> int:
        """Returns the efficiency quotient."""
        return int(self._state.attributes[SERVICE_EFFICIENCY]['efficiencyQuotient'])

    @property
    @backend_parameter
    def last_trip_list(self) -> List[LastTrip]:
        """Returns the list of last trips."""
        ret_list = []
        last_trip_list = self._state.attributes[SERVICE_EFFICIENCY].get('lastTripList', [])
        for trip in last_trip_list:
            ret_list.append(LastTrip(trip))
        return ret_list

    @property
    @backend_parameter
    def life_time_list(self) -> List[LifeTime]:
        """Returns the life time list."""
        ret_list = []
        last_time_list = self._state.attributes[SERVICE_EFFICIENCY].get('lifeTimeList', [])
        for time in last_time_list:
            ret_list.append(LifeTime(time))
        return ret_list

    @property
    @backend_parameter
    def characteristic_list(self) -> List[Characteristic]:
        """Returns the characteristic list."""
        ret_list = []
        characteristic_list = self._state.attributes[SERVICE_EFFICIENCY].get('characteristicList', [])
        for characteristic in characteristic_list:
            ret_list.append(Characteristic(characteristic))
        return ret_list



"""Mock for Connected Drive Backend."""

import re
import os
import json
from typing import List
from bimmer_connected.country_selector import Regions

RESPONSE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'responses')

TEST_USERNAME = 'some_user'
TEST_PASSWORD = 'my_secret'
TEST_REGION = Regions.REST_OF_WORLD
G31_VIN = 'G31_NBTevo_VIN'
F48_VIN = 'F48_VIN'
I01_VIN = 'I01_VIN'
F15_VIN = 'F15_VIN'
I01_NOREX_VIN = 'I01_NOREX_VIN'
F45_VIN = 'F45_VIN'
F31_VIN = 'F31_VIN'
G30_PHEV_OS7_VIN = 'G30_PHEV_OS7_VIN'

#: Mapping of VINs to test data directories
TEST_VEHICLE_DATA = {
    G31_VIN: 'G31_NBTevo',
    F48_VIN: 'F48',
    I01_VIN: 'I01_REX',
    I01_NOREX_VIN: 'I01_NOREX',
    F15_VIN: 'F15',
    F45_VIN: 'F45',
    F31_VIN: 'F31',
    G30_PHEV_OS7_VIN: 'G30_PHEV_OS7'
}

_AUTH_RESPONSE_HEADERS = {
    'X-Powered-By': 'JOY',
    'Content-Type': 'application/json',
    'X-CorrelationID': 'Id-cde5a45a0802dd010000000031546948 0',
    'Via': '1.0 lpb2vcn02 (BMW Group API Gateway)',
    'Transfer-Encoding': 'chunked',
    'X-NodeID': '02',
    'Max-Forwards': '20',
    'Date': 'Sun, 11 Mar 2018 08:16:13 GMT',
    'Content-Encoding': 'gzip',
    'Location': ('https://www.bmw-connecteddrive.com/app/static/external-dispatch.html'
                 '#access_token=TOKEN&token_type=Bearer&expires_in=7199')
}

# VehicleState has different names than the json file. So we need to map some of the
# parameters.
ATTRIBUTE_MAPPING = {
    'remainingFuel': 'remaining_fuel',
    'position': 'gps_position',
    'cbsData': 'condition_based_services',
    'checkControlMessages': 'check_control_messages',
    'doorLockState': 'door_lock_state',
    'updateReason': 'last_update_reason',
    'chargingLevelHv': 'charging_level_hv',
    'chargingStatus': 'charging_status',
    'maxRangeElectric': 'max_range_electric',
    'remainingRangeElectric': 'remaining_range_electric',
    'parkingLight': 'parking_lights',
    'remainingRangeFuel': 'remaining_range_fuel',
    'updateTime': 'timestamp',
    'chargingTimeRemaining': 'charging_time_remaining'
}

# these are additional attributes in the API, not available in the status.json
ADDITIONAL_ATTRIBUTES = [
    'remaining_range_total',    # added by bimmer_connected
    'charging_time_remaining',  # only present while charging
    'sunroof',                  # not available in all vehicles
    'lids',                     # required for existing Home Assistant binary sensors
    'windows',                  # required for existing Home Assistant binary sensors
    'lights_parking',           # required for existing Home Assistant binary sensors
    'steering',                 # para not available in all vehicles
]

# there attributes are not (yet) implemented
MISSING_ATTRIBUTES = [
    'remainingRangeFuelMls',      # we're not using miles
    'remainingRangeElectricMls',  # we're not using miles
    'maxRangeElectricMls',        # we're not using miles
    'chargingTimeRemaining',      # only present while charging
    'sunroof',                    # not available in all vehicles
    'lights_parking',             # required for existing Home Assistant binary sensors
    'steering',                   # para not available in all vehicles
    'vehicleCountry',             # para not available in all vehicles
]

AVAILABLE_STATES_MAPPING = {
    "statisticsAvailable": {True: ["LAST_TRIP", "ALL_TRIPS"]},
    "chargingControl": {"WEEKLY_PLANNER": ["CHARGING_PROFILE"]},
    "lastDestinations": {"SUPPORTED": ["DESTINATIONS"]},
    "rangeMap": {"RANGE_CIRCLE": ["RANGEMAP"]}
}

POI_DATA = {
    "lat": 37.4028943,
    "lon": -121.9700289,
    "name": "49ers",
    "additional_info": "Hi Sam",
    "street": "4949 Marie P DeBartolo Way",
    "city": "Santa Clara",
    "postal_code": "CA 95054",
    "country": "United States",
    "website": "https://www.49ers.com/",
    "phone_numbers": ["+1 408-562-4949"]
}

POI_REQUEST = {
    "min": ("data=%7B%22poi%22%3A+%7B%22lat%22%3A+37.4028943%2C+%22lon%22%3A+-121.9700289"
            "%2C+%22additionalInfo%22%3A+%22Sent+with+%5Cu2665+by+bimmer_connected%22%7D%7D"),
    "all": ("data=%7B%22poi%22%3A+%7B%22lat%22%3A+37.4028943%2C+%22lon%22%3A+-121.9700289"
            "%2C+%22name%22%3A+%2249ers%22%2C+%22additionalInfo%22%3A+%22Hi+Sam%22"
            "%2C+%22street%22%3A+%224949+Marie+P+DeBartolo+Way%22%2C+%22city%22%3A+%22Santa+Clara%22"
            "%2C+%22postalCode%22%3A+%22CA+95054%22%2C+%22country%22%3A+%22United+States%22"
            "%2C+%22website%22%3A+%22https%3A%2F%2Fwww.49ers.com%2F%22"
            "%2C+%22phoneNumbers%22%3A+%5B%22%2B1+408-562-4949%22%5D%7D%7D")
}

MESSAGE_DATA = {"subject": "This is a subject...", "text": "This is a message!"}

MESSAGE_REQUEST = {
    "min": ("data=%7B%22poi%22%3A+%7B%22additionalInfo%22%3A+%22This+is+a+message%21%22%7D%7D"),
    "all": ("data=%7B%22poi%22%3A+%7B%22name%22%3A+%22This+is+a+subject...%22%2C+%22additionalInfo"
            "%22%3A+%22This+is+a+message%21%22%7D%7D")
}


def load_response_json(filename: str) -> dict:
    """load a stored response from a file"""
    with open(os.path.join(RESPONSE_DIR, filename)) as json_file:
        return json.load(json_file)


class BackendMock:
    """Mock for Connected Drive Backend."""

    # pylint: disable=too-many-arguments

    def __init__(self) -> None:
        """Constructor."""
        self.last_request = []
        self.responses = [
            MockResponse('https://.+/gcdm/.*/?oauth/token',
                         headers=_AUTH_RESPONSE_HEADERS,
                         data_files=['G31_NBTevo/auth_response.json'],
                         status_code=200),
            MockResponse('https://.+/webapi/v1/user/vehicles$',
                         data_files=['vehicles.json']),
        ]

    def get(self, url: str, headers: dict = None, data: str = None, allow_redirects: bool = None, params=None) \
            -> 'MockResponse':
        """Mock for requests.get function."""
        self.last_request.append(MockRequest(url, headers, data, request_type='GET', allow_redirects=allow_redirects,
                                             params=params))
        return self._find_response(url)

    def post(self, url: str, headers: dict = None, data: str = None, allow_redirects: bool = None, params=None) \
            -> 'MockResponse':
        """Mock for requests.post function."""
        self.last_request.append(MockRequest(url, headers, data, request_type='GET', allow_redirects=allow_redirects,
                                             params=params))
        return self._find_response(url)

    def add_response(self, regex: str, data: str = None, data_files: List[str] = None,
                     headers: dict = None, status_code=200) -> None:
        """Add a response to the backend."""
        self.responses.append(MockResponse(regex, data, data_files, headers, status_code))

    def _find_response(self, url) -> 'MockResponse':
        """Find a proper response for a requested url."""
        for response in self.responses:
            if response.regex.search(url):
                response.step_responses()
                return response
        return MockResponse(regex='', data='unknown url: {}'.format(url), status_code=404)

    def setup_default_vehicles(self) -> None:
        """Setup the vehicle configuration in a mock backend."""
        for vin, path in TEST_VEHICLE_DATA.items():
            self.add_response('https://.+/webapi/v1/user/vehicles/{vin}/status$'.format(vin=vin),
                              data_files=['{path}/status.json'.format(path=path)])


class MockRequest:  # pylint: disable=too-few-public-methods
    """Stores the attributes of a request."""

    # pylint: disable=too-many-arguments

    def __init__(self, url, headers, data, request_type=None, allow_redirects=None, params=None) -> None:
        self.url = url
        self.headers = headers
        self.data = data
        self.request_type = request_type
        self.allow_redirects = allow_redirects
        self.params = params


class MockResponse:
    """Mocks requests.response."""

    # pylint: disable=too-many-arguments

    def __init__(self, regex: str, data: str = None, data_files: List[str] = None, headers: dict = None,
                 status_code: int = 200) -> None:
        """Constructor."""
        self.regex = re.compile(regex)
        self.status_code = status_code
        self.headers = headers
        self._usage_count = 0
        if self.headers is None:
            self.headers = dict()

        if data_files is not None:
            self._data = []
            for data_file in data_files:
                with open(os.path.join(RESPONSE_DIR, data_file)) as response:
                    self._data.append(response.read())
        else:
            self._data = [data]

    def json(self) -> dict:
        """Parse the text of the response as a jsons string."""
        return json.loads(self.text)

    @property
    def text(self) -> str:
        """Get the raw data from the response."""
        return self._data[self._usage_count-1]

    def step_responses(self):
        """Step through the list of responses.

        The last response will be repeated forever.
        """
        self._usage_count = min(len(self._data), self._usage_count+1)



#!/usr/bin/python3
"""Simple executable to demonstrate and test the usage of the library."""

import argparse
import logging
import json
import time
import sys

from pathlib import Path

import requests

from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.country_selector import get_region_from_name, valid_regions
from bimmer_connected.vehicle import VehicleViewDirection

TEXT_VIN = 'Vehicle Identification Number'


def main_parser() -> argparse.ArgumentParser:
    """Creates the ArgumentParser with all relevant subparsers."""
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser(description='A simple executable to use and test the library.')
    subparsers = parser.add_subparsers(dest='cmd')
    subparsers.required = True

    status_parser = subparsers.add_parser('status', description='Get the current status of the vehicle.')
    _add_default_arguments(status_parser)
    _add_position_arguments(status_parser)

    fingerprint_parser = subparsers.add_parser('fingerprint', description='Save a vehicle fingerprint.')
    _add_default_arguments(fingerprint_parser)
    _add_position_arguments(fingerprint_parser)
    fingerprint_parser.set_defaults(func=fingerprint)

    flash_parser = subparsers.add_parser('lightflash', description='Flash the vehicle lights.')
    _add_default_arguments(flash_parser)
    flash_parser.add_argument('vin', help=TEXT_VIN)
    flash_parser.set_defaults(func=light_flash)

    finder_parser = subparsers.add_parser('vehiclefinder', description='Update the vehicle GPS location.')
    _add_default_arguments(finder_parser)
    finder_parser.add_argument('vin', help=TEXT_VIN)
    finder_parser.set_defaults(func=vehicle_finder)

    image_parser = subparsers.add_parser('image', description='Download a vehicle image.')
    _add_default_arguments(image_parser)
    image_parser.add_argument('vin', help=TEXT_VIN)
    image_parser.set_defaults(func=image)

    sendpoi_parser = subparsers.add_parser('sendpoi', description='Send a point of interest to the vehicle.')
    _add_default_arguments(sendpoi_parser)
    sendpoi_parser.add_argument('vin', help=TEXT_VIN)
    sendpoi_parser.add_argument('latitude', help='Latitude of the POI', type=float)
    sendpoi_parser.add_argument('longitude', help='Longitude of the POI', type=float)
    sendpoi_parser.add_argument('--name', help='(optional, display only) Name of the POI', nargs='?', default=None)
    sendpoi_parser.add_argument('--street', help='(optional, display only) Street & House No. of the POI',
                                nargs='?', default=None)
    sendpoi_parser.add_argument('--city', help='(optional, display only) City of the POI', nargs='?', default=None)
    sendpoi_parser.add_argument('--postalcode', help='(optional, display only) Postal code of the POI',
                                nargs='?', default=None)
    sendpoi_parser.add_argument('--country', help='(optional, display only) Country of the POI',
                                nargs='?', default=None)
    sendpoi_parser.set_defaults(func=send_poi)

    sendpoi_from_address_parser = subparsers.add_parser('sendpoi_from_address',
                                                        description=('Send a point of interest parsed from a'
                                                                     ' street address to the vehicle.'))
    _add_default_arguments(sendpoi_from_address_parser)
    sendpoi_from_address_parser.add_argument('vin', help=TEXT_VIN)
    sendpoi_from_address_parser.add_argument('-n', '--name', help='(optional, display only) Name of the POI',
                                             nargs='?', default=None)
    sendpoi_from_address_parser.add_argument('-a', '--address', nargs='+',
                                             help="Address (e.g. 'Street 17, city, zip, country')")
    sendpoi_from_address_parser.set_defaults(func=send_poi_from_address)

    message_parser = subparsers.add_parser('sendmessage', description='Send a text message to the vehicle.')
    _add_default_arguments(message_parser)
    message_parser.add_argument('vin', help=TEXT_VIN)
    message_parser.add_argument('text', help='Text to be sent.')
    message_parser.add_argument('subject', help='(optional) Message subject', nargs='?')
    message_parser.set_defaults(func=send_message)

    return parser


def get_status(args) -> None:
    """Get the vehicle status."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    if args.lat and args.lng:
        for vehicle in account.vehicles:
            vehicle.set_observer_position(args.lat, args.lng)
    account.update_vehicle_states()

    print('Found {} vehicles: {}'.format(
        len(account.vehicles),
        ','.join([v.name for v in account.vehicles])))

    for vehicle in account.vehicles:
        print('VIN: {}'.format(vehicle.vin))
        print('Mileage: {}'.format(vehicle.state.vehicle_status.mileage))
        print('Vehicle properties:')
        print(json.dumps(vehicle.attributes, indent=4))
        print('Vehicle status:')
        print(json.dumps(vehicle.state.vehicle_status.attributes, indent=4))


def fingerprint(args) -> None:
    """Save the vehicle fingerprint."""
    time_dir = Path.home() / 'vehicle_fingerprint' / time.strftime("%Y-%m-%d_%H-%M-%S")
    time_dir.mkdir(parents=True)

    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region),
                                    log_responses=time_dir)
    account.set_observer_position(args.lat, args.lng)
    if args.lat and args.lng:
        for vehicle in account.vehicles:
            vehicle.set_observer_position(args.lat, args.lng)
    account.update_vehicle_states()

    print('fingerprint of the vehicles written to {}'.format(time_dir))


def light_flash(args) -> None:
    """Trigger the vehicle to flash its lights."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)
    if not vehicle:
        valid_vins = ", ".join(v.vin for v in account.vehicles)
        print('Error: Could not find vehicle for VIN "{}". Valid VINs are: {}'.format(args.vin, valid_vins))
        return
    status = vehicle.remote_services.trigger_remote_light_flash()
    print(status.state)


def vehicle_finder(args) -> None:
    """Trigger the vehicle finder to locate it."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)
    if not vehicle:
        valid_vins = ", ".join(v.vin for v in account.vehicles)
        print('Error: Could not find vehicle for VIN "{}". Valid VINs are: {}'.format(args.vin, valid_vins))
        return
    status = vehicle.remote_services.trigger_remote_vehicle_finder()
    print(status.state)


def image(args) -> None:
    """Download a rendered image of the vehicle."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)

    with open('image.png', 'wb') as output_file:
        image_data = vehicle.get_vehicle_image(400, 400, VehicleViewDirection.FRONT)
        output_file.write(image_data)
    print('vehicle image saved to image.png')


def send_poi(args) -> None:
    """Send Point Of Interest to car."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)
    poi_data = dict(
        lat=args.latitude,
        lon=args.longitude,
        name=args.name,
        street=args.street,
        city=args.city,
        postal_code=args.postalcode,
        country=args.country
    )
    vehicle.remote_services.trigger_send_poi(poi_data)


def send_poi_from_address(args) -> None:
    """Create Point of Interest from OSM Nominatim and send to car."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)
    address = [(str(' '.join(args.address)))]
    try:
        response = requests.get("https://nominatim.openstreetmap.org",
                                params={
                                    "q": address,
                                    "format": "json",
                                    "addressdetails": 1,
                                    "limit": 1
                                }).json()[0]
    except IndexError:
        print('\nAddress not found')
        sys.exit(1)
    address = response.get("address", {})
    city = address.get("city")
    town = address.get("town")

    poi_data = dict(
        lat=response["lat"],
        lon=response["lon"],
        name=args.name,
        street=address.get("road"),
        city=town if city is None and town is not None else None,
        postal_code=address.get("postcode"),
        country=address.get("country")
    )
    vehicle.remote_services.trigger_send_poi(poi_data)


def send_message(args) -> None:
    """Send a message to car."""
    account = ConnectedDriveAccount(args.username, args.password, get_region_from_name(args.region))
    vehicle = account.get_vehicle(args.vin)
    msg_data = dict(
        text=args.text,
        subject=args.subject
    )
    vehicle.remote_services.trigger_send_message(msg_data)


def _add_default_arguments(parser: argparse.ArgumentParser):
    """Add the default arguments username, password, region to the parser."""
    parser.add_argument('username', help='Connected Drive username')
    parser.add_argument('password', help='Connected Drive password')
    parser.add_argument('region', choices=valid_regions(), help='Region of the Connected Drive account')


def _add_position_arguments(parser: argparse.ArgumentParser):
    """Add the lat and lng attributes to the parser."""
    parser.add_argument('lat', type=float, nargs='?', const=0.0,
                        help='(optional) Your current GPS latitude (as float)')
    parser.add_argument('lng', type=float, nargs='?', const=0.0,
                        help='(optional) Your current GPS longitude (as float)')
    parser.set_defaults(func=get_status)


def main():
    """Main function."""
    parser = main_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()



"""Models the charging profiles of a vehicle."""

import logging
from typing import List
from enum import Enum

from bimmer_connected.const import SERVICE_CHARGING_PROFILE

_LOGGER = logging.getLogger(__name__)


class ChargingMode(Enum):
    """Charging mode of electric vehicle."""
    IMMEDIATE_CHARGING = 'IMMEDIATE_CHARGING'
    DELAYED_CHARGING = 'DELAYED_CHARGING'


class ChargingPreferences(Enum):
    """Charging preferences of electric vehicle."""
    NO_PRESELECTION = 'NO_PRESELECTION'
    CHARGING_WINDOW = 'CHARGING_WINDOW'


class TimerTypes(Enum):
    """Different Timer-Types."""
    TIMER_1 = 'timer1'
    TIMER_2 = 'timer2'
    TIMER_3 = 'timer3'
    OVERRIDE_TIMER = 'overrideTimer'


class ChargingWindow:
    """
    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, window_dict: dict):
        self._window_dict = window_dict

    @property
    def start_time(self) -> str:
        """Start of the charging window."""
        return self._window_dict["startTime"]

    @property
    def end_time(self) -> str:
        """End of the charging window."""
        return self._window_dict["endTime"]


class ClimatizationTimer:
    """
    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, clima_dict: dict):
        self._clima_dict = clima_dict

    @property
    def departure_time(self) -> str:
        """Deperture time for this timer."""
        return self._clima_dict["departureTime"]

    @property
    def timer_enabled(self) -> bool:
        """Is the timer enabled."""
        return self._clima_dict["timerEnabled"]

    @property
    def weekdays(self) -> List[str]:
        """Active weekdays for this timer."""
        return self._clima_dict["weekdays"]


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'ChargingProfile', *args, **kwargs):
        # pylint: disable=protected-access
        if self._state.attributes[SERVICE_CHARGING_PROFILE] is None:
            raise ValueError('No data available for vehicles charging profile!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class ChargingProfile:  # pylint: disable=too-many-public-methods
    """Models the charging profile of a vehicle."""

    def __init__(self, state):
        """Constructor."""
        self._state = state

    @property
    @backend_parameter
    def attributes(self) -> dict:
        """Retrieve all attributes from the sever.

        This does not parse the results in any way.
        """
        return self._state.attributes[SERVICE_CHARGING_PROFILE]

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self._state.attributes[SERVICE_CHARGING_PROFILE][item]

    @property
    @backend_parameter
    def is_pre_entry_climatization_enabled(self) -> bool:
        """Get status of pre-entry climatization."""
        return bool(self._state.attributes[SERVICE_CHARGING_PROFILE]['climatizationEnabled'])

    @property
    @backend_parameter
    def pre_entry_climatization_timer(self) -> dict:
        """List of pre entry climatization timer messages."""
        timer_list = {}
        for timer in TimerTypes:
            try:
                timer_list[timer] = ClimatizationTimer(self._state.attributes[SERVICE_CHARGING_PROFILE][timer.value])
            except KeyError:
                _LOGGER.debug('Timer %s not found', timer.value)
        return timer_list

    @property
    @backend_parameter
    def prefered_charging_window(self) -> ChargingWindow:
        """Returns the prefered charging window."""
        return ChargingWindow(self._state.attributes[SERVICE_CHARGING_PROFILE]['preferredChargingWindow'])

    @property
    @backend_parameter
    def charging_preferences(self) -> str:
        """Returns the prefered charging preferences."""
        return ChargingPreferences(self._state.attributes[SERVICE_CHARGING_PROFILE]['chargingPreferences'])

    @property
    @backend_parameter
    def charging_mode(self) -> str:
        """Returns the prefered charging mode."""
        return ChargingMode(self._state.attributes[SERVICE_CHARGING_PROFILE]['chargingMode'])



"""Test for old deprecated state api."""

import unittest
from unittest import mock
import datetime
from test import load_response_json, TEST_REGION, TEST_PASSWORD, TEST_USERNAME, BackendMock, G31_VIN, \
    ATTRIBUTE_MAPPING, MISSING_ATTRIBUTES, F48_VIN
from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.state import VehicleState, LockState, ParkingLightState, ChargingState
from bimmer_connected.const import SERVICE_STATUS
from bimmer_connected.vehicle_status import ConditionBasedServiceStatus, LidState

G31_TEST_DATA = load_response_json('G31_NBTevo/status.json')
G31_NO_POSITION_TEST_DATA = load_response_json('G31_NBTevo/status_position_disabled.json')
F48_TEST_DATA = load_response_json('F48/status.json')
I01_TEST_DATA = load_response_json('I01_REX/status.json')


class TestState(unittest.TestCase):
    """Test for VehicleState."""

    # pylint: disable=protected-access

    def test_parse_g31(self):
        """Test if the parsing of the attributes is working."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = G31_TEST_DATA['vehicleStatus']

        self.assertEqual(4126, state.mileage)

        zone = datetime.timezone(datetime.timedelta(0, 3600))
        self.assertEqual(datetime.datetime(year=2018, month=3, day=10, hour=11, minute=39, second=41, tzinfo=zone),
                         state.timestamp)

        self.assertTrue(state.is_vehicle_tracking_enabled)
        self.assertAlmostEqual(50.5050, state.gps_position[0])
        self.assertAlmostEqual(10.1010, state.gps_position[1])

        self.assertAlmostEqual(33, state.remaining_fuel)

        self.assertAlmostEqual(321, state.remaining_range_fuel)
        self.assertAlmostEqual(state.remaining_range_fuel, state.remaining_range_total)

        self.assertEqual('VEHICLE_SHUTDOWN', state.last_update_reason)

        cbs = state.condition_based_services
        self.assertEqual(3, len(cbs))
        self.assertEqual(ConditionBasedServiceStatus.OK, cbs[0].state)
        self.assertEqual(datetime.datetime(year=2020, month=1, day=1), cbs[0].due_date)
        self.assertEqual(25000, cbs[0].due_distance)

        self.assertEqual(ConditionBasedServiceStatus.OK, cbs[1].state)
        self.assertEqual(datetime.datetime(year=2022, month=1, day=1), cbs[1].due_date)
        self.assertEqual(60000, cbs[1].due_distance)

        self.assertTrue(state.are_all_cbs_ok)

        self.assertFalse(state.are_parking_lights_on)
        self.assertEqual(ParkingLightState.OFF, state.parking_lights)

    def test_parse_g31_no_psoition(self):
        """Test parsing of G31 data with position tracking disabled in the vehicle."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = G31_NO_POSITION_TEST_DATA['vehicleStatus']

        self.assertFalse(state.is_vehicle_tracking_enabled)
        self.assertIsNone(state.gps_position)

    def test_parse_f48(self):
        """Test if the parsing of the attributes is working."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = F48_TEST_DATA['vehicleStatus']

        self.assertEqual(21529, state.mileage)

        zone = datetime.timezone(datetime.timedelta(0, 3600))
        self.assertEqual(datetime.datetime(year=2018, month=3, day=10, hour=19, minute=35, second=30, tzinfo=zone),
                         state.timestamp)

        self.assertTrue(state.is_vehicle_tracking_enabled)
        self.assertAlmostEqual(50.505050, state.gps_position[0])
        self.assertAlmostEqual(10.1010101, state.gps_position[1])

        self.assertAlmostEqual(39, state.remaining_fuel)

        self.assertAlmostEqual(590, state.remaining_range_fuel)

        self.assertEqual('DOOR_STATE_CHANGED', state.last_update_reason)

        cbs = state.condition_based_services
        self.assertEqual(3, len(cbs))
        self.assertEqual(ConditionBasedServiceStatus.OK, cbs[0].state)
        self.assertEqual(datetime.datetime(year=2019, month=7, day=1), cbs[0].due_date)
        self.assertEqual(9000, cbs[0].due_distance)

        self.assertEqual(ConditionBasedServiceStatus.OK, cbs[1].state)
        self.assertEqual(datetime.datetime(year=2021, month=7, day=1), cbs[1].due_date)
        self.assertEqual(39000, cbs[1].due_distance)

        self.assertTrue(state.are_all_cbs_ok)

        self.assertFalse(state.are_parking_lights_on)
        self.assertEqual(ParkingLightState.OFF, state.parking_lights)

    def test_parse_i01(self):
        """Test if the parsing of the attributes is working."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = I01_TEST_DATA['vehicleStatus']

        self.assertEqual(48, state.remaining_range_electric)
        self.assertEqual(154, state.remaining_range_total)
        self.assertEqual(94, state.max_range_electric)
        self.assertEqual(ChargingState.CHARGING, state.charging_status)
        self.assertEqual(datetime.timedelta(minutes=332), state.charging_time_remaining)
        self.assertEqual(54, state.charging_level_hv)

    def test_parse_timeformat(self):
        """Test parsing of the time string."""
        date = "2018-03-10T11:39:41+0100"
        zone = datetime.timezone(datetime.timedelta(0, 3600))
        self.assertEqual(datetime.datetime(year=2018, month=3, day=10, hour=11, minute=39, second=41, tzinfo=zone),
                         VehicleState._parse_datetime(date))

    def test_missing_attribute(self):
        """Test if error handling is working correctly."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = dict()
        self.assertIsNone(state.mileage)

    @mock.patch('bimmer_connected.vehicle.VehicleState.update_data')
    def test_no_attributes(self, _):
        """Test if error handling is working correctly."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        self.assertIsNone(state.mileage)

    def test_update_data(self):
        """Test update_data method."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
            vehicle = account.get_vehicle(G31_VIN)
            vehicle.state.update_data()

            backend_mock.setup_default_vehicles()

            vehicle.state.update_data()
            self.assertEqual(4126, vehicle.state.mileage)

    def test_lids(self):
        """Test features around lids."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = G31_TEST_DATA['vehicleStatus']

        for lid in state.lids:
            self.assertEqual(LidState.CLOSED, lid.state)

        self.assertEqual(6, len(list(state.lids)))
        self.assertEqual(0, len(list(state.open_lids)))
        self.assertTrue(state.all_lids_closed)

        state._attributes[SERVICE_STATUS]['doorDriverFront'] = LidState.OPEN
        self.assertFalse(state.all_lids_closed)

    def test_windows_g31(self):
        """Test features around lids."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = G31_TEST_DATA['vehicleStatus']

        for window in state.vehicle_status.windows:
            self.assertEqual(LidState.CLOSED, window.state)

        self.assertEqual(5, len(list(state.windows)))
        self.assertEqual(0, len(list(state.open_windows)))
        self.assertTrue(state.all_windows_closed)

        state._attributes[SERVICE_STATUS]['windowDriverFront'] = LidState.INTERMEDIATE
        self.assertFalse(state.all_windows_closed)

    def test_windows_f48(self):
        """Test features around lids."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = F48_TEST_DATA['vehicleStatus']

        for window in state.windows:
            self.assertEqual(LidState.CLOSED, window.state)

        self.assertEqual(4, len(list(state.windows)))

    def test_door_locks(self):
        """Test the door locks."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = G31_TEST_DATA['vehicleStatus']

        self.assertEqual(LockState.SECURED, state.door_lock_state)

    def test_parsing_attributes(self):
        """Test parsing different attributes of the vehicle.

        Just make sure parsing that no exception is raised and we get not-None values.
        """
        backend_mock = BackendMock()
        # list of attributes that are ignored at the moment
        ignored_attributes = [ATTRIBUTE_MAPPING.get(a, a) for a in MISSING_ATTRIBUTES]
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            backend_mock.setup_default_vehicles()
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
            account.update_vehicle_states()

            for vehicle in account.vehicles:
                print(vehicle.name)
                for attribute in (a for a in vehicle.available_attributes if a not in ignored_attributes):
                    self.assertIsNotNone(getattr(vehicle.state, attribute), attribute)

    def test_check_control_messages(self):
        """Test handling of check control messages.

        F48 is the only vehicle with active Check Control Messages, so we only expect to get something there.
        """
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            backend_mock.setup_default_vehicles()
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
            account.update_vehicle_states()

            for vehicle in account.vehicles:
                print(vehicle.name, vehicle.vin)
                if vehicle.vin == F48_VIN:
                    self.assertTrue(vehicle.state.has_check_control_messages)
                else:
                    self.assertFalse(vehicle.state.has_check_control_messages)

    def test_ccm_f48(self):
        """Test parsing of a check control message."""
        account = unittest.mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_STATUS] = F48_TEST_DATA['vehicleStatus']

        ccms = state.vehicle_status.check_control_messages
        self.assertEqual(1, len(ccms))
        ccm = ccms[0]
        self.assertEqual(955, ccm["ccmId"])
        self.assertEqual(41544, ccm["ccmMileage"])
        self.assertIn("Tyre pressure", ccm["ccmDescriptionShort"])
        self.assertIn("continue driving", ccm["ccmDescriptionLong"])



"""Models the last trip of a vehicle."""

import logging

from bimmer_connected.const import SERVICE_LAST_TRIP

_LOGGER = logging.getLogger(__name__)


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'LastTrip', *args, **kwargs):
        # pylint: disable=protected-access
        if self._state.attributes[SERVICE_LAST_TRIP] is None:
            raise ValueError('No data available for vehicle last trips!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class LastTrip:  # pylint: disable=too-many-public-methods
    """Models the last trip of a vehicle."""

    def __init__(self, state):
        """Constructor."""
        self._state = state

    @property
    @backend_parameter
    def attributes(self) -> dict:
        """Retrieve all attributes from the sever.

        This does not parse the results in any way.
        """
        return self._state.attributes[SERVICE_LAST_TRIP]

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self._state.attributes[SERVICE_LAST_TRIP][item]

    @property
    @backend_parameter
    def efficiency_value(self) -> float:
        """Returns the effenciency value."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['efficiencyValue'])

    @property
    @backend_parameter
    def total_distance(self) -> float:
        """Returns the total distance."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['totalDistance'])

    @property
    @backend_parameter
    def electric_distance(self) -> float:
        """Returns the electric distance."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['electricDistance'])

    @property
    @backend_parameter
    def average_electric_consumption(self) -> float:
        """Returns the average electric consumption."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['avgElectricConsumption'])

    @property
    @backend_parameter
    def average_recuperation(self) -> float:
        """Returns the average recuperation."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['avgRecuperation'])

    @property
    @backend_parameter
    def driving_mode_value(self) -> float:
        """Returns the driving mode value."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['drivingModeValue'])

    @property
    @backend_parameter
    def acceleration_value(self) -> float:
        """Returns the acceleration value."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['accelerationValue'])

    @property
    @backend_parameter
    def anticipation_value(self) -> float:
        """Returns the anticipation value."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['anticipationValue'])

    @property
    @backend_parameter
    def total_consumption_value(self) -> float:
        """Returns the total consumption value."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['totalConsumptionValue'])

    @property
    @backend_parameter
    def auxiliary_consumption_value(self) -> float:
        """Returns the auxiliary consumption value."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['auxiliaryConsumptionValue'])

    @property
    @backend_parameter
    def average_combined_consumption(self) -> float:
        """Returns the average combined consumption."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['avgCombinedConsumption'])

    @property
    @backend_parameter
    def electric_distance_ratio(self) -> float:
        """Returns the electric distance ratio."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['electricDistanceRatio'])

    @property
    @backend_parameter
    def saved_fuel(self) -> float:
        """Returns the saved fuel."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['savedFuel'])

    @property
    @backend_parameter
    def date(self) -> str:
        """Returns the date."""
        return self._state.attributes[SERVICE_LAST_TRIP]['date']

    @property
    @backend_parameter
    def duration(self) -> float:
        """Returns the duration."""
        return float(self._state.attributes[SERVICE_LAST_TRIP]['duration'])



"""Models the all trips of a vehicle."""

import logging

from bimmer_connected.const import SERVICE_ALL_TRIPS

_LOGGER = logging.getLogger(__name__)


def backend_parameter_statistic(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'StatisticValues', *args, **kwargs):
        # pylint: disable=protected-access
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class StatisticValues:
    """
    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, stats_dict: dict):
        self._stats_dict = stats_dict

    @property
    @backend_parameter_statistic
    def community_low(self) -> float:
        """Get the community low value."""
        return float(self._stats_dict["communityLow"])

    @property
    @backend_parameter_statistic
    def community_average(self) -> float:
        """Get the community average value."""
        return float(self._stats_dict["communityAverage"])

    @property
    @backend_parameter_statistic
    def community_high(self) -> float:
        """Get the community high value."""
        return float(self._stats_dict["communityHigh"])

    @property
    @backend_parameter_statistic
    def user_average(self) -> float:
        """Get the user average value."""
        return float(self._stats_dict["userAverage"])

    @property
    @backend_parameter_statistic
    def user_high(self) -> float:
        """Get the user high value."""
        return float(self._stats_dict["userHigh"])

    @property
    @backend_parameter_statistic
    def user_total(self) -> float:
        """Get the user total value."""
        return float(self._stats_dict["userTotal"])

    @property
    @backend_parameter_statistic
    def user_current_charge_cycle(self) -> float:
        """Get the users current charge cycle."""
        return float(self._stats_dict["userCurrentChargeCycle"])


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'AllTrips', *args, **kwargs):
        # pylint: disable=protected-access
        if self._state.attributes[SERVICE_ALL_TRIPS] is None:
            raise ValueError('No data available for vehicles trips!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class AllTrips:  # pylint: disable=too-many-public-methods
    """Models the all trips service of a vehicle."""

    def __init__(self, state):
        """Constructor."""
        self._state = state

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self._state.attributes[SERVICE_ALL_TRIPS][item]

    @property
    @backend_parameter
    def reset_date(self) -> str:
        """Returns the reset date."""
        return self._state.attributes[SERVICE_ALL_TRIPS]['resetDate']

    @property
    @backend_parameter
    def battery_size_max(self) -> int:
        """Returns the maximal battery size, in Wh."""
        return int(self._state.attributes[SERVICE_ALL_TRIPS]['batterySizeMax'])

    @property
    @backend_parameter
    def saved_co2(self) -> int:
        """Returns the saved CO2, in kg."""
        return float(self._state.attributes[SERVICE_ALL_TRIPS]['savedCO2'])

    @property
    @backend_parameter
    def saved_co2_green_energy(self) -> int:
        """Returns the save CO2GreenEnergym in kg."""
        return float(self._state.attributes[SERVICE_ALL_TRIPS]['savedCO2greenEnergy'])

    @property
    @backend_parameter
    def total_saved_fuel(self) -> int:
        """Returns the total saved fuel, in l."""
        return float(self._state.attributes[SERVICE_ALL_TRIPS]['totalSavedFuel'])

    @property
    @backend_parameter
    def average_electric_consumption(self) -> StatisticValues:
        """Returns the average electric consumption."""
        return StatisticValues(self._state.attributes[SERVICE_ALL_TRIPS]['avgElectricConsumption'])

    @property
    @backend_parameter
    def average_recopuration(self) -> StatisticValues:
        """Returns the average recopuration."""
        return StatisticValues(self._state.attributes[SERVICE_ALL_TRIPS]['avgRecuperation'])

    @property
    @backend_parameter
    def chargecycle_range(self) -> StatisticValues:
        """Returns the charge cycle range."""
        return StatisticValues(self._state.attributes[SERVICE_ALL_TRIPS]['chargecycleRange'])

    @property
    @backend_parameter
    def total_electric_distance(self) -> StatisticValues:
        """Returns the total electric distance."""
        return StatisticValues(self._state.attributes[SERVICE_ALL_TRIPS]['totalElectricDistance'])

    @property
    @backend_parameter
    def average_combined_consumption(self) -> StatisticValues:
        """Returns the average combined consumption."""
        return StatisticValues(self._state.attributes[SERVICE_ALL_TRIPS]['avgCombinedConsumption'])



"""Models the last destinations of a vehicle."""

import logging
from typing import List
from enum import Enum

from bimmer_connected.const import SERVICE_DESTINATIONS

_LOGGER = logging.getLogger(__name__)


class DestinationType(Enum):
    """Types of destinations."""
    DESTINATION = 'DESTINATION'


class Destination:
    """
    This class provides a nicer API than parsing the JSON format directly.
    """

    def __init__(self, dest_dict: dict):
        self._dest_dict = dest_dict

    @property
    def latitude(self) -> float:
        """latitude of this destination."""
        return float(self._dest_dict["lat"])

    @property
    def longitude(self) -> float:
        """longitude of this destination."""
        return float(self._dest_dict["lon"])

    @property
    def country(self) -> str:
        """Country of this destination."""
        return self._dest_dict["country"]

    @property
    def city(self) -> str:
        """City of this destination."""
        return self._dest_dict["city"]

    @property
    def street(self) -> str:
        """Street of this destination."""
        return self._dest_dict["street"]

    @property
    def destination_type(self) -> DestinationType:
        """Type of this destination."""
        return DestinationType(self._dest_dict["type"])

    @property
    def created_at(self) -> str:
        """Date of creation of this destination."""
        return self._dest_dict["createdAt"]


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'LastDestinations', *args, **kwargs):
        # pylint: disable=protected-access
        if self._state.attributes[SERVICE_DESTINATIONS] is None:
            raise ValueError('No data available for vehicle destinations!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class LastDestinations:  # pylint: disable=too-many-public-methods
    """Models the last destinations of a vehicle."""

    def __init__(self, state):
        """Constructor."""
        self._state = state

    @property
    @backend_parameter
    def attributes(self) -> dict:
        """Retrieve all attributes from the sever.

        This does not parse the results in any way.
        """
        return self._state.attributes[SERVICE_DESTINATIONS]

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self._state.attributes[SERVICE_DESTINATIONS][item]

    @property
    @backend_parameter
    def last_destinations(self) -> List[Destination]:
        """Get the list of last destinations."""
        destinations_list = []
        for dest in self._state.attributes[SERVICE_DESTINATIONS]:
            destinations_list.append(Destination(dest))
        return destinations_list



"""URLs for different services and error code mapping."""

AUTH_URL = 'https://{gcdm_oauth_endpoint}/oauth/token'
BASE_URL = 'https://{server}/webapi/v1'
BASE_URL_LEGACY = 'https://{server}/api/vehicle'

VEHICLES_URL = BASE_URL + '/user/vehicles'
VEHICLE_VIN_URL = VEHICLES_URL + '/{vin}'
VEHICLE_STATUS_URL = VEHICLE_VIN_URL + '/status'
REMOTE_SERVICE_STATUS_URL = VEHICLE_VIN_URL + '/serviceExecutionStatus?serviceType={service_type}'
REMOTE_SERVICE_URL = VEHICLE_VIN_URL + "/executeService"
VEHICLE_IMAGE_URL = VEHICLE_VIN_URL + "/image?width={width}&height={height}&view={view}"
VEHICLE_POI_URL = VEHICLE_VIN_URL + '/sendpoi'
VEHICLE_STATISTICS_URL = VEHICLE_VIN_URL + '/statistics'
VEHICLE_STATISTICS_LAST_TRIP_URL = VEHICLE_STATISTICS_URL + '/lastTrip'
VEHICLE_STATISTICS_ALL_TRIPS_URL = VEHICLE_STATISTICS_URL + '/allTrips'
VEHICLE_CHARGING_PROFILE_URL = VEHICLE_VIN_URL + '/chargingprofile'
VEHICLE_DESTINATIONS_URL = VEHICLE_VIN_URL + '/destinations'
VEHICLE_RANGEMAP_URL = VEHICLE_VIN_URL + '/rangemap'
VEHICLE_EFFICIENCY = BASE_URL_LEGACY + '/efficiency' + '/v1/{vin}'
VEHICLE_NAVIGATION = BASE_URL_LEGACY + '/navigation' + '/v1/{vin}'

SERVICE_STATUS = 'STATUS'
SERVICE_LAST_TRIP = 'LAST_TRIP'
SERVICE_ALL_TRIPS = 'ALL_TRIPS'
SERVICE_CHARGING_PROFILE = 'CHARGING_PROFILE'
SERVICE_DESTINATIONS = 'DESTINATIONS'
SERVICE_RANGEMAP = 'RANGEMAP'
SERVICE_EFFICIENCY = 'EFFICIENCY'
SERVICE_NAVIGATION = 'NAVIGATION'

# Possible error codes, other codes are mapped to UNKNOWN_ERROR
ERROR_CODE_MAPPING = {
    401: 'UNAUTHORIZED',
    404: 'NOT_FOUND',
    405: 'MOBILE_ACCESS_DISABLED',
    408: 'VEHICLE_UNAVAILABLE',
    423: 'ACCOUNT_LOCKED',
    429: 'TOO_MANY_REQUESTS',
    500: 'SERVER_ERROR',
    503: 'SERVICE_MAINTENANCE',
}



"""Models the state of a vehicle."""

import logging
import datetime
from typing import List

from bimmer_connected.const import SERVICE_STATUS, VEHICLE_STATUS_URL, SERVICE_LAST_TRIP, \
    VEHICLE_STATISTICS_LAST_TRIP_URL, SERVICE_ALL_TRIPS, VEHICLE_STATISTICS_ALL_TRIPS_URL, \
    SERVICE_CHARGING_PROFILE, VEHICLE_CHARGING_PROFILE_URL, SERVICE_DESTINATIONS, VEHICLE_DESTINATIONS_URL, \
    SERVICE_RANGEMAP, VEHICLE_RANGEMAP_URL, SERVICE_EFFICIENCY, VEHICLE_EFFICIENCY, SERVICE_NAVIGATION, \
    VEHICLE_NAVIGATION

from bimmer_connected.vehicle_status import VehicleStatus, LockState, ParkingLightState, ChargingState, \
    CheckControlMessage, ConditionBasedServiceReport, Lid, Window
from bimmer_connected.last_trip import LastTrip
from bimmer_connected.all_trips import AllTrips
from bimmer_connected.charging_profile import ChargingProfile
from bimmer_connected.last_destinations import LastDestinations
from bimmer_connected.range_maps import RangeMaps
from bimmer_connected.navigation import Navigation
from bimmer_connected.efficiency import Efficiency

_LOGGER = logging.getLogger(__name__)


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'VehicleState', *args, **kwargs):
        # pylint: disable=protected-access
        if self._attributes is None:
            raise ValueError('No data available for vehicles state!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class VehicleState:
    """Models the state of a vehicle."""

    # pylint: disable=too-many-public-methods
    # pylint: disable=too-many-instance-attributes
    # Nine is reasonable in this case.
    def __init__(self, account, vehicle):
        """Constructor."""
        self._account = account
        self._vehicle = vehicle
        self._attributes = {}
        self.vehicle_status = VehicleStatus(self)
        self.all_trips = AllTrips(self)
        self.charging_profile = ChargingProfile(self)
        self.last_trip = LastTrip(self)
        self.last_destinations = LastDestinations(self)
        self.range_maps = RangeMaps(self)
        self.navigation = Navigation(self)
        self.efficiency = Efficiency(self)

        self._url = {
            SERVICE_STATUS: VEHICLE_STATUS_URL,
            SERVICE_LAST_TRIP: VEHICLE_STATISTICS_LAST_TRIP_URL,
            SERVICE_ALL_TRIPS: VEHICLE_STATISTICS_ALL_TRIPS_URL,
            SERVICE_CHARGING_PROFILE: VEHICLE_CHARGING_PROFILE_URL,
            SERVICE_DESTINATIONS: VEHICLE_DESTINATIONS_URL,
            SERVICE_RANGEMAP: VEHICLE_RANGEMAP_URL,
            SERVICE_EFFICIENCY: VEHICLE_EFFICIENCY,
            SERVICE_NAVIGATION: VEHICLE_NAVIGATION}

        self._key = {
            SERVICE_STATUS: 'vehicleStatus',
            SERVICE_LAST_TRIP: 'lastTrip',
            SERVICE_ALL_TRIPS: 'allTrips',
            SERVICE_CHARGING_PROFILE: 'weeklyPlanner',
            SERVICE_DESTINATIONS: 'destinations',
            SERVICE_RANGEMAP: 'rangemap',
            SERVICE_EFFICIENCY: '',
            SERVICE_NAVIGATION: ''}

        for service in self._url:
            self._attributes[service] = {}

    def update_data(self) -> None:
        """Read new status data from the server."""
        _LOGGER.debug('requesting new data from connected drive')
        format_string = '%Y-%m-%dT%H:%M:%S'
        timestamp = datetime.datetime.now().strftime(format_string)
        params = {
            'deviceTime': timestamp,
            'dlat': self._vehicle.observer_latitude,
            'dlon': self._vehicle.observer_longitude,
        }

        for service in self._vehicle.available_state_services:
            try:
                response = self._account.send_request(
                    self._url[service].format(server=self._account.server_url, vin=self._vehicle.vin),
                    logfilename=service, params=params)
                if not self._key[service]:
                    self._attributes[service] = response.json()
                else:
                    self._attributes[service] = response.json()[self._key[service]]
            except IOError:
                _LOGGER.debug('Service %s failed', service)
            except KeyError:  # When JSON contains no service-key
                _LOGGER.debug('Service %s failed', service)

        _LOGGER.debug(self._attributes)
        _LOGGER.debug('received new data from connected drive')

    @property
    @backend_parameter
    def attributes(self) -> dict:
        """Retrieve all attributes from the sever.

        This does not parse the results in any way.
        """
        return self._attributes

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self.vehicle_status.attributes[item]

    @staticmethod
    def _parse_datetime(date_str: str) -> datetime.datetime:
        """Convert a time string into datetime."""
        date_format = "%Y-%m-%dT%H:%M:%S%z"
        return datetime.datetime.strptime(date_str, date_format)

    @property
    @backend_parameter
    def timestamp(self) -> datetime.datetime:
        """DEPRECATED: Use state.vehicle_status.timestamp instead.

        Get the timestamp when the data was recorded."""
        return self.vehicle_status.timestamp

    @property
    @backend_parameter
    def gps_position(self) -> (float, float):
        """DEPRECATED: Use state.vehicle_status.gps_position instead.

        Get the last known position of the vehicle.

        Returns a tuple of (latitude, longitude).
        This only provides data, if the vehicle tracking is enabled!
        """
        return self.vehicle_status.gps_position

    @property
    @backend_parameter
    def is_vehicle_tracking_enabled(self) -> bool:
        """DEPRECATED: Use state.vehicle_status.is_vehicle_tracking_enabled instead.

        Check if the position tracking of the vehicle is enabled.

        The server return "OK" if tracking is enabled and "DRIVER_DISABLED" if it is disabled in the vehicle.
        """
        return bool(self.vehicle_status.is_vehicle_tracking_enabled)

    @property
    @backend_parameter
    def mileage(self) -> int:
        """DEPRECATED: Use state.vehicle_status.mileage instead.

        Get the mileage of the vehicle.

        Returns a tuple of (value, unit_of_measurement)
        """
        return self.vehicle_status.mileage

    @property
    @backend_parameter
    def remaining_range_fuel(self) -> int:
        """DEPRECATED: Use state.vehicle_status.remaining_range_fuel instead.

        Get the remaining range of the vehicle on fuel.

        Returns a tuple of (value, unit_of_measurement)
        """
        return self.vehicle_status.remaining_range_fuel

    @property
    @backend_parameter
    def remaining_fuel(self) -> int:
        """DEPRECATED: Use state.vehicle_status.remaining_fuel instead.

        Get the remaining fuel of the vehicle.

        Returns a tuple of (value, unit_of_measurement)
        """
        return self.vehicle_status.remaining_fuel

    @property
    @backend_parameter
    def lids(self) -> List[Lid]:
        """DEPRECATED: Use state.vehicle_status.lids instead.

        Get all lids (doors+hatch+trunk) of the car.
        """
        return self.vehicle_status.lids

    @property
    @backend_parameter
    def open_lids(self) -> List[Lid]:
        """DEPRECATED: Use state.vehicle_status.open_lids instead.

        Get all open lids of the car.
        """
        return self.vehicle_status.open_lids

    @property
    @backend_parameter
    def all_lids_closed(self) -> bool:
        """DEPRECATED: Use state.vehicle_status.all_lids_closed instead.

        Check if all lids are closed.
        """
        return self.vehicle_status.all_lids_closed

    @property
    @backend_parameter
    def windows(self) -> List[Window]:
        """DEPRECATED: Use state.vehicle_status.windows instead.

        Get all windows (doors+sun roof) of the car.
        """
        return self.vehicle_status.windows

    @property
    @backend_parameter
    def open_windows(self) -> List[Window]:
        """DEPRECATED: Use state.vehicle_status.open_windows instead.

        Get all open windows of the car.
        """
        return self.vehicle_status.open_windows

    @property
    @backend_parameter
    def all_windows_closed(self) -> bool:
        """DEPRECATED: Use state.vehicle_status.all_windows_closed instead.

        Check if all windows are closed.
        """
        return self.vehicle_status.all_windows_closed

    @property
    @backend_parameter
    def door_lock_state(self) -> LockState:
        """DEPRECATED: Use state.vehicle_status.all_windows_closed instead.

        Get state of the door locks.
        """
        return self.vehicle_status.door_lock_state

    @property
    @backend_parameter
    def last_update_reason(self) -> str:
        """DEPRECATED: Use state.vehicle_status.last_update_reason instead.

        The reason for the last state update"""
        return self.vehicle_status.last_update_reason

    @property
    @backend_parameter
    def last_charging_end_result(self) -> str:
        """DEPRECATED: Use state.vehicle_status.last_charging_end_result instead.

        Get the last charging end result"""
        return self.vehicle_status.last_charging_end_result

    @property
    @backend_parameter
    def connection_status(self) -> str:
        """DEPRECATED: Use state.vehicle_status.connection_status instead.

        Get status of the connection"""
        return self.vehicle_status.connection_status

    @property
    @backend_parameter
    def condition_based_services(self) -> List[ConditionBasedServiceReport]:
        """DEPRECATED: Use state.vehicle_status.condition_based_services instead.

        Get status of the condition based services.
        """
        return self.vehicle_status.condition_based_services

    @property
    @backend_parameter
    def are_all_cbs_ok(self) -> bool:
        """DEPRECATED: Use state.vehicle_status.are_all_cbs_ok instead.

        Check if the status of all condition based services is "OK".
        """
        return bool(self.vehicle_status.are_all_cbs_ok)

    @property
    @backend_parameter
    def parking_lights(self) -> ParkingLightState:
        """DEPRECATED: Use state.vehicle_status.parking_lights instead.

        Get status of parking lights.

        :returns None if status is unknown.
        """
        return self.vehicle_status.parking_lights

    @property
    @backend_parameter
    def are_parking_lights_on(self) -> bool:
        """DEPRECATED: Use state.vehicle_status.are_parking_lights_on instead.

        Get status of parking lights.

        :returns None if status is unknown.
        """
        return bool(self.vehicle_status.are_parking_lights_on)

    @property
    @backend_parameter
    def remaining_range_electric(self) -> int:
        """DEPRECATED: Use state.vehicle_status.remaining_range_electric instead.

        Remaining range on battery, in kilometers.
        """
        return self.vehicle_status.remaining_range_electric

    @property
    @backend_parameter
    def remaining_range_total(self) -> int:
        """DEPRECATED: Use state.vehicle_status.remaining_range_total instead.

        Get the total remaining range of the vehicle in kilometers.
        That is electrical range + fuel range.
        """
        return self.vehicle_status.remaining_range_total

    @property
    @backend_parameter
    def max_range_electric(self) -> int:
        """DEPRECATED: Use state.vehicle_status.max_range_electric instead.

        Maximum range on battery, in kilometers.
        This can change with driving style and temperature.
        """
        return self.vehicle_status.max_range_electric

    @property
    @backend_parameter
    def charging_status(self) -> ChargingState:
        """DEPRECATED: Use state.vehicle_status.charging_status instead.

        Charging state of the vehicle.
        """
        return self.vehicle_status.charging_status

    @property
    @backend_parameter
    def charging_time_remaining(self) -> datetime.timedelta:
        """DEPRECATED: Use state.vehicle_status.charging_time_remaining instead.

        Get the remaining charging time.
        """
        return self.vehicle_status.charging_time_remaining

    @property
    @backend_parameter
    def charging_level_hv(self) -> int:
        """DEPRECATED: Use state.vehicle_status.charging_level_hv instead.

        State of charge of the high voltage battery in percent.
        """
        return self.vehicle_status.charging_level_hv

    @property
    @backend_parameter
    def check_control_messages(self) -> List[CheckControlMessage]:
        """DEPRECATED: Use state.vehicle_status.check_control_messages instead.

        List of check control messages.
        """
        return self.vehicle_status.check_control_messages

    @property
    @backend_parameter
    def has_check_control_messages(self) -> bool:
        """DEPRECATED: Use state.vehicle_status.has_check_control_messages instead.

        Return true if any check control message is present.
        """
        return bool(self.vehicle_status.has_check_control_messages)



"""Tests for ConnectedDriveAccount."""
import json
import unittest
from unittest import mock
from test import BackendMock, G31_VIN, TEST_USERNAME, TEST_PASSWORD, TEST_REGION
from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.country_selector import Regions


class TestAccount(unittest.TestCase):
    """Tests for ConnectedDriveAccount."""

    # pylint: disable=protected-access

    def test_token_vehicles(self):
        """Test getting backend token and vehicle list."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, Regions.REST_OF_WORLD)
            self.assertIsNotNone(account._oauth_token)
            self.assertEqual(9, len(account.vehicles))
            vehicle = account.get_vehicle(G31_VIN)
            self.assertEqual(G31_VIN, vehicle.vin)

            self.assertIsNone(account.get_vehicle('invalid_vin'))

    def test_invalid_send_response(self):
        """Test parsing the results of an invalid request"""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, Regions.REST_OF_WORLD)
            with self.assertRaises(IOError):
                account.send_request('invalid_url')

    def test_invalid_auth(self):
        """Test if the host is set correctly in the request."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            with mock.patch('bimmer_connected.account.get_gcdm_oauth_endpoint') as mocked_endpoint:
                mocked_endpoint().return_value = 'customer.bmwgroup.com/gcdm/invalid'
                with self.assertRaises(OSError):
                    ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, Regions.REST_OF_WORLD)

    def test_china_header(self):
        """Test if the host is set correctly in the request."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, Regions.CHINA)
            request = [r for r in backend_mock.last_request if 'oauth' in r.url][0]
            self.assertEqual('customer.bmwgroup.cn', request.headers['Host'])

    def test_anonymize_data(self):
        """Test anonymization function."""
        test_dict = {
            'vin': 'secret',
            'a sub-dict': {
                'lat': 666,
                'lon': 666,
                'heading': 666,
            },
            'licensePlate': 'secret',
            'public': 'public_data',
            'a_list': [
                {'vin': 'secret'},
                {
                    'lon': 666,
                    'public': 'more_public_data',
                }
            ],
            'b_list': ['a', 'b'],
            'empty_list': [],
        }
        anon_text = json.dumps(ConnectedDriveAccount._anonymize_data(test_dict))
        self.assertNotIn('secret', anon_text)
        self.assertNotIn('666', anon_text)
        self.assertIn('public_data', anon_text)
        self.assertIn('more_public_data', anon_text)

    def test_vehicle_search_case(self):
        """Check if the search for the vehicle by VIN is NOT case sensitive."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

        vin = account.vehicles[1].vin
        self.assertEqual(vin, account.get_vehicle(vin).vin)
        self.assertEqual(vin, account.get_vehicle(vin.lower()).vin)
        self.assertEqual(vin, account.get_vehicle(vin.upper()).vin)

    def test_set_observer_value(self):
        """Test set_observer_position with valid arguments."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, Regions.REST_OF_WORLD)

            account.set_observer_position(1.0, 2.0)
            for vehicle in account.vehicles:
                self.assertEqual(vehicle.observer_latitude, 1.0)
                self.assertEqual(vehicle.observer_longitude, 2.0)

    def test_set_observer_not_set(self):
        """Test set_observer_position with no arguments."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, Regions.REST_OF_WORLD)

            for vehicle in account.vehicles:
                self.assertEqual(vehicle.observer_latitude, 0.0)
                self.assertEqual(vehicle.observer_longitude, 0.0)

            account.set_observer_position(None, None)

            for vehicle in account.vehicles:
                self.assertEqual(vehicle.observer_latitude, 0.0)
                self.assertEqual(vehicle.observer_longitude, 0.0)

    def test_set_observer_some_none(self):
        """Test set_observer_position with invalid arguments."""
        backend_mock = BackendMock()
        with mock.patch('bimmer_connected.account.requests', new=backend_mock):
            account = ConnectedDriveAccount(TEST_USERNAME, TEST_PASSWORD, Regions.REST_OF_WORLD)

            with self.assertRaises(ValueError):
                account.set_observer_position(None, 2.0)

            with self.assertRaises(ValueError):
                account.set_observer_position(1.0, None)



"""Library to read data from the BMW Connected Drive portal.

The library bimmer_connected provides a Python interface to interact
with the BMW Connected Drive web service. It allows you to read
the current state of the vehicle and also trigger remote services.

Disclaimer:
This library is not affiliated with or endorsed by BMW Group.
"""

import datetime
import logging
import pathlib
import urllib
import json
from threading import Lock
from typing import Callable, List
import requests

from bimmer_connected.country_selector import Regions, get_server_url, get_gcdm_oauth_endpoint
from bimmer_connected.vehicle import ConnectedDriveVehicle
from bimmer_connected.const import AUTH_URL, VEHICLES_URL, ERROR_CODE_MAPPING

_LOGGER = logging.getLogger(__name__)


class ConnectedDriveAccount:  # pylint: disable=too-many-instance-attributes
    """Create a new connection to the BMW Connected Drive web service.

    :param username: Connected drive user name
    :param password: Connected drive password
    :param country: Country for which the account was created. For a list of valid countries,
                check https://www.bmw-connecteddrive.com .
                Use the name of the countries exactly as on the website.
    :param log_responses: If log_responses is set, all responses from the server will
                be loged into this directory. This can be used for later analysis of the different
                responses for different vehicles.
    :param retries_on_500_error: If retries_on_500_error is set, a communication with the
                Connected Drive server will automatically be retried the number of times
                specified in the event the error code received was 500. This sometimes
                occurs (presumably) due to bugs in the server implementation.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, username: str, password: str, region: Regions, log_responses: pathlib.Path = None,
                 retries_on_500_error: int = 5) -> None:
        self._region = region
        self._server_url = None
        self._username = username
        self._password = password
        self._oauth_token = None
        self._refresh_token = None
        self._token_expiration = None
        self._log_responses = log_responses
        self._retries_on_500_error = retries_on_500_error
        #: list of vehicles associated with this account.
        self._vehicles = []
        self._lock = Lock()
        self._update_listeners = []

        self._get_vehicles()

    def _get_oauth_token(self) -> None:
        """Get a new auth token from the server."""
        with self._lock:
            if self._token_expiration is not None and datetime.datetime.now() < self._token_expiration:
                _LOGGER.debug('Old token is still valid. Not getting a new one.')
                return

            _LOGGER.debug('getting new oauth token')
            url = AUTH_URL.format(
                gcdm_oauth_endpoint=get_gcdm_oauth_endpoint(self._region)
            )

            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "Content-Length": "124",
                "Connection": "Keep-Alive",
                "Host": urllib.parse.urlparse(url).netloc,
                "Accept-Encoding": "gzip",
                "Authorization": "Basic ZDc2NmI1MzctYTY1NC00Y2JkLWEzZGMtMGNhNTY3MmQ3ZjhkOjE1"
                                 "ZjY5N2Y2LWE1ZDUtNGNhZC05OWQ5LTNhMTViYzdmMzk3Mw==",
                "Credentials": "nQv6CqtxJuXWP74xf3CJwUEP:1zDHx6un4cDjybLENN3kyfumX2kEYigWPcQpdvDRpIBk7rOJ",
                "User-Agent": "okhttp/3.12.2",
            }

            # we really need all of these parameters
            values = {
                'scope': 'authenticate_user vehicle_data remote_services',
                'grant_type': 'password',
                'username': self._username,
                'password': self._password,
            }

            data = urllib.parse.urlencode(values)
            expected_response_code = 200
            try:
                response = self.send_request(url, data=data, headers=headers, allow_redirects=False,
                                             expected_response=expected_response_code, post=True)
            except OSError as exception:
                msg = 'Authentication failed. Maybe your password is invalid?'
                _LOGGER.error(msg)
                _LOGGER.exception(exception)
                raise OSError(msg) from exception

            response_json = response.json()

            self._oauth_token = response_json['access_token']
            expiration_time = int(response_json['expires_in'])
            self._token_expiration = datetime.datetime.now() + datetime.timedelta(seconds=expiration_time)
            _LOGGER.debug('got new token %s with expiration date %s', self._oauth_token, self._token_expiration)

    @property
    def request_header(self):
        """Generate a header for HTTP requests to the server."""
        self._get_oauth_token()
        headers = {
            "accept": "application/json",
            "Authorization": "Bearer {}".format(self._oauth_token),
            "referer": "https://www.bmw-connecteddrive.de/app/index.html",
        }
        return headers

    def send_request(self, url: str, data=None, headers=None, expected_response=200, post=False, allow_redirects=True,
                     logfilename: str = None, params: dict = None):
        """Send an http request to the server.

        If the http headers are not set, default headers are generated.
        You can choose if you want a GET or POST request.
        """
        if headers is None:
            headers = self.request_header

        for i in range(self._retries_on_500_error + 1):
            if post:
                response = requests.post(url, headers=headers, data=data, allow_redirects=allow_redirects,
                                         params=params)
            else:
                response = requests.get(url, headers=headers, data=data, allow_redirects=allow_redirects,
                                        params=params)

            if response.status_code != expected_response:
                if response.status_code == 500:
                    _LOGGER.debug("Error 500 on attempt %d", i+1)
                    continue

                error_description = ERROR_CODE_MAPPING.get(response.status_code, "UNKNOWN_ERROR")
                msg = ("The BMW Connected Drive portal returned an error: {} (received status code {} and expected {})."
                       .format(error_description, response.status_code, expected_response))
                _LOGGER.debug(msg)
                _LOGGER.debug(response.text)
                raise IOError(msg)
            break

        self._log_response_to_file(response, logfilename)
        return response

    def _log_response_to_file(self, response: requests.Response, logfilename: str = None) -> None:
        """If a log path is set, log all resonses to a file."""
        if self._log_responses is None or logfilename is None:
            return

        anonymized_data = json.dumps(self._anonymize_data(response.json()), indent=2, sort_keys=True)

        output_path = None
        count = 0

        while output_path is None or output_path.exists():
            output_path = self._log_responses / '{}_{}.txt'.format(logfilename, count)
            count += 1

        with open(output_path, 'w') as logfile:
            logfile.write(anonymized_data)

    @staticmethod
    def _anonymize_data(json_data: dict) -> dict:
        """Replace parts of the logfiles containing personal information."""

        replacements = {
            'lat': 12.3456,
            'lon': 34.5678,
            'heading': 123,
            'vin': 'some_vin',
            'licensePlate': 'some_license_plate',
            'name': 'some_name',
            'city': 'some_city',
            'street': 'some_street',
            'streetNumber': '999',
            'postalCode': 'some_postal_code',
            'phone': 'some_phone',
        }

        if isinstance(json_data, list):
            json_data = [ConnectedDriveAccount._anonymize_data(v) for v in json_data]
        elif isinstance(json_data, dict):
            for key, value in json_data.items():
                if key in replacements:
                    json_data[key] = replacements[key]
                else:
                    json_data[key] = ConnectedDriveAccount._anonymize_data(value)

        return json_data

    @property
    def server_url(self) -> str:
        """Get the url of the server for this country."""
        if self._server_url is None:
            self._server_url = get_server_url(self._region)
        return self._server_url

    def _get_vehicles(self):
        """Retrieve list of vehicle for the account."""
        _LOGGER.debug('Getting vehicle list')
        self._get_oauth_token()
        response = self.send_request(VEHICLES_URL.format(server=self.server_url), headers=self.request_header,
                                     logfilename='vehicles')

        for vehicle_dict in response.json()['vehicles']:
            self._vehicles.append(ConnectedDriveVehicle(self, vehicle_dict))

    def get_vehicle(self, vin: str) -> ConnectedDriveVehicle:
        """Get vehicle with given VIN.

        The search is NOT case sensitive.
        :param vin: VIN of the vehicle you want to get.
        :return: Returns None if no such vehicle is found.
        """
        for car in self.vehicles:
            if car.vin.upper() == vin.upper():
                return car
        return None

    def update_vehicle_states(self) -> None:
        """Update the state of all vehicles.

        Notify all listeners of the vehicle state update.
        """
        for car in self.vehicles:
            car.update_state()
        for listener in self._update_listeners:
            listener()

    def add_update_listener(self, listener: Callable) -> None:
        """Add a listener for state updates."""
        self._update_listeners.append(listener)

    @property
    def vehicles(self) -> List[ConnectedDriveVehicle]:
        """Get list of vehicle of this account"""
        return self._vehicles

    def __str__(self):
        """Use the user name as id for the account class."""
        return '{}: {}'.format(self.__class__, self._username)

    def set_observer_position(self, latitude: float, longitude: float) -> None:
        """Set the position of the observer for all vehicles.

        see VehicleViewDirection.set_observer_position() for more details.
        """
        if bool(latitude) != bool(longitude):
            raise ValueError('Either latitude and longitude are both not None or both are None.')
        if latitude and longitude:
            for vehicle in self._vehicles:
                vehicle.set_observer_position(latitude, longitude)



"""Test the country selection class."""
import unittest

from bimmer_connected.country_selector import valid_regions, Regions, get_region_from_name


class TestCountrySelector(unittest.TestCase):
    """Test the country selection class."""

    def test_valid_regions(self):
        """Test getting list of regions."""
        self.assertIn('china', valid_regions())

    def test_region_from_name(self):
        """Test parsing region from string."""
        self.assertEqual(Regions.CHINA, get_region_from_name('China'))
        self.assertEqual(Regions.REST_OF_WORLD, get_region_from_name('rest_of_world'))
        self.assertEqual(Regions.NORTH_AMERICA, get_region_from_name('nOrTh_AmErica'))

    def test_invalid_region(self):
        """Test exception handling."""
        with self.assertRaises(ValueError):
            get_region_from_name('random_text')



"""Trigger remote services on a vehicle."""

import datetime
import json
import logging
import time
from enum import Enum
from urllib.parse import urlencode

import requests

from bimmer_connected.const import (REMOTE_SERVICE_STATUS_URL,
                                    REMOTE_SERVICE_URL,
                                    VEHICLE_POI_URL)

TIME_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'

_LOGGER = logging.getLogger(__name__)

#: time in seconds between polling updates on the status of a remote service
_POLLING_CYCLE = 1

#: maximum number of seconds to wait for the server to return a positive answer
_POLLING_TIMEOUT = 120

#: time in seconds to wait before updating the vehicle state from the server
_UPDATE_AFTER_REMOTE_SERVICE_DELAY = 10


class ExecutionState(Enum):
    """Enumeration of possible states of the execution of a remote service."""
    INITIATED = 'INITIATED'
    PENDING = 'PENDING'
    DELIVERED = 'DELIVERED'
    EXECUTED = 'EXECUTED'
    UNKNOWN = 'UNKNOWN'


class _Services(Enum):
    """Enumeration of possible services to be executed."""
    REMOTE_LIGHT_FLASH = 'LIGHT_FLASH'
    REMOTE_VEHICLE_FINDER = 'VEHICLE_FINDER'
    REMOTE_DOOR_LOCK = 'DOOR_LOCK'
    REMOTE_DOOR_UNLOCK = 'DOOR_UNLOCK'
    REMOTE_HORN = 'HORN_BLOW'
    REMOTE_AIR_CONDITIONING = 'CLIMATE_NOW'


# pylint: disable=too-many-instance-attributes, too-few-public-methods
class PointOfInterest:
    """Point of interest to be sent to the vehicle.

    The latitude/longitude of a POI are mandatory, all other attributes are optional. CamelCase attribute names are
    used here so that we do not have to convert the names between the attributes and the keys as expected on the server.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, lat: float, lon: float, name: str = None,
                 additional_info: str = None, street: str = None, city: str = None,
                 postal_code: str = None, country: str = None, website: str = None,
                 phone_numbers: [str] = None):
        """Create a PointOfInterest with attributes in camelCase as required by the API.

        :arg lat: latitude of the POI
        :arg lon: longitude of the POI
        :arg name: name of the POI (Optional)
        :arg additional_info: additional text shown below the address (Optional)
        :arg street: street with house number of the POI (Optional)
        :arg city: city of the POI (Optional)
        :arg postal_code: zip code of the POI (Optional)
        :arg country: country of the POI (Optional)
        :arg website: website of the POI (Optional)
        :arg phone_numbers: List of phone numbers of the POI (Optional)
        """
        # pylint: disable=invalid-name
        self.lat = lat  # type: float
        self.lon = lon  # type: float
        self.name = name  # type: str
        self.additionalInfo = additional_info if additional_info is not None \
            else 'Sent with ♥ by bimmer_connected'  # type: str
        self.street = street  # type: str
        self.city = city  # type: str
        self.postalCode = postal_code  # type: str
        self.country = country  # type: str
        self.website = website  # type: str
        self.phoneNumbers = phone_numbers  # type: list[str]


class Message:
    """Text message or PointOfInterst to be sent to the vehicle."""

    @classmethod
    def from_poi(cls, poi: PointOfInterest):
        """Create a message from a PointOfInterest"""
        return cls(poi.__dict__)

    @classmethod
    def from_text(cls, text: str, subject: str = None):
        """Create a text message"""
        return cls({"name": subject, "additionalInfo": text[:255]})

    def __init__(self, data: dict):
        self.data = data

    @property
    def as_server_request(self) -> str:
        """Convert to a dictionary so that it can be sent to the server."""
        result = {
            'poi': {k: v for k, v in self.data.items() if v is not None}
        }
        return urlencode({'data': json.dumps(result)})


class RemoteServiceStatus:  # pylint: disable=too-few-public-methods
    """Wraps the status of the execution of a remote service."""

    def __init__(self, response: dict):
        """Construct a new object from a dict."""
        status = response['executionStatus']
        self.state = ExecutionState(status.get('status', 'UNKNOWN'))
        self.event_id = status['eventId']

    @staticmethod
    def _parse_timestamp(timestamp: str) -> datetime.datetime:
        """Parse the timestamp format from the response."""
        offset = int(timestamp[-3:])
        time_zone = datetime.timezone(datetime.timedelta(hours=offset))
        result = datetime.datetime.strptime(timestamp[:-3], TIME_FORMAT)
        result.replace(tzinfo=time_zone)
        return result


class RemoteServices:
    """Trigger remote services on a vehicle."""

    def __init__(self, account, vehicle):
        """Constructor."""
        self._account = account
        self._vehicle = vehicle

    def trigger_remote_light_flash(self) -> RemoteServiceStatus:
        """Trigger the vehicle to flash its headlights.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Triggering remote light flash')
        # needs to be called via POST, GET is not working
        self._trigger_remote_service(_Services.REMOTE_LIGHT_FLASH, post=True)
        return self._block_until_done(_Services.REMOTE_LIGHT_FLASH)

    def trigger_remote_door_lock(self) -> RemoteServiceStatus:
        """Trigger the vehicle to lock its doors.

        A state update is triggered after this, as the lock state of the vehicle changes.
        """
        _LOGGER.debug('Triggering remote door lock')
        # needs to be called via POST, GET is not working
        self._trigger_remote_service(_Services.REMOTE_DOOR_LOCK, post=True)
        result = self._block_until_done(_Services.REMOTE_DOOR_LOCK)
        self._trigger_state_update()
        return result

    def trigger_remote_door_unlock(self) -> RemoteServiceStatus:
        """Trigger the vehicle to unlock its doors.

        A state update is triggered after this, as the lock state of the vehicle changes.
        """
        _LOGGER.debug('Triggering remote door unlock')
        # needs to be called via POST, GET is not working
        self._trigger_remote_service(_Services.REMOTE_DOOR_UNLOCK, post=True)
        result = self._block_until_done(_Services.REMOTE_DOOR_UNLOCK)
        self._trigger_state_update()
        return result

    def trigger_remote_horn(self) -> RemoteServiceStatus:
        """Trigger the vehicle to sound its horn.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Triggering remote horn sound')
        # needs to be called via POST, GET is not working
        self._trigger_remote_service(_Services.REMOTE_HORN, post=True)
        return self._block_until_done(_Services.REMOTE_HORN)

    def trigger_remote_air_conditioning(self) -> RemoteServiceStatus:
        """Trigger the air conditioning to start.

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Triggering remote air conditioning')
        # needs to be called via POST, GET is not working
        self._trigger_remote_service(_Services.REMOTE_AIR_CONDITIONING, post=True)
        result = self._block_until_done(_Services.REMOTE_AIR_CONDITIONING)
        self._trigger_state_update()
        return result

    def _trigger_remote_service(self, service_id: _Services, post=False) -> requests.Response:
        """Trigger a generic remote service.

        You can choose if you want a POST or a GET operation.
        """
        data = {'serviceType': service_id.value}
        url = REMOTE_SERVICE_URL.format(vin=self._vehicle.vin, server=self._account.server_url)

        return self._account.send_request(url, post=post, data=data)

    def _block_until_done(self, service: _Services) -> RemoteServiceStatus:
        """Keep polling the server until we get a final answer.

        :raises IOError: if there is no final answer before _POLLING_TIMEOUT
        """
        fail_after = datetime.datetime.now() + datetime.timedelta(seconds=_POLLING_TIMEOUT)
        while True:
            status = self._get_remote_service_status(service)
            _LOGGER.debug('current state if remote service is: %s', status.state.value)
            if status.state not in [ExecutionState.UNKNOWN, ExecutionState.PENDING, ExecutionState.DELIVERED]:
                return status
            if datetime.datetime.now() > fail_after:
                raise IOError(
                    'Timeout on getting final answer from server. Current state: {}'.format(status.state.value))
            time.sleep(_POLLING_CYCLE)

    def _get_remote_service_status(self, service: _Services) -> RemoteServiceStatus:
        """The the execution status of the last remote service that was triggered.

        As the status changes over time, you probably need to poll this.
        Recommended polling time is AT LEAST one second as the reaction is sometimes quite slow.
        """
        _LOGGER.debug('getting remote service status')
        url = REMOTE_SERVICE_STATUS_URL.format(
            server=self._account.server_url,
            vin=self._vehicle.vin,
            service_type=service.value)
        response = self._account.send_request(url)
        try:
            json_result = response.json()
            return RemoteServiceStatus(json_result)
        except ValueError:
            _LOGGER.error('Error decoding json response from the server.')
            _LOGGER.debug(response.headers)
            _LOGGER.debug(response.text)
            raise

    def _trigger_state_update(self) -> None:
        time.sleep(_UPDATE_AFTER_REMOTE_SERVICE_DELAY)
        self._account.update_vehicle_states()

    def trigger_send_message(self, data: dict) -> RemoteServiceStatus:
        """Send a message to the vehicle.

        :param data: A dictonary containing a 'text' and an optional 'subject'
        :type data: dict

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Sending message to car')
        if "text" not in data:
            raise TypeError("from_text() missing 1 required positional argument: 'text'")
        self._send_message(Message.from_text(data['text'], data.get('subject')))
        # _send_message has no separate ExecutionStates
        return RemoteServiceStatus({'executionStatus': {'status': 'EXECUTED', 'eventId': -1}})

    def trigger_send_poi(self, data: dict) -> RemoteServiceStatus:
        """Send a PointOfInterest to the vehicle.

        :param data: A dictonary containing at least 'lat' and 'lon' and optionally
            'name', 'additionalInfo', 'street', 'city', 'postalCode', 'country',
            'website' or 'phoneNumbers'
        :type data: dict

        A state update is NOT triggered after this, as the vehicle state is unchanged.
        """
        _LOGGER.debug('Sending PointOfInterest to car')
        if "lat" not in data or "lon" not in data:
            raise TypeError("__init__() missing 2 required positional arguments: 'lat' and 'lon'")
        poi = PointOfInterest(**data)
        self._send_message(Message.from_poi(poi))
        # _send_message has no separate ExecutionStates
        return RemoteServiceStatus({'executionStatus': {'status': 'EXECUTED', 'eventId': -1}})

    def _send_message(self, msg: Message) -> None:
        """Send a message/point of interest to the vehicle."""
        url = VEHICLE_POI_URL.format(
            vin=self._vehicle.vin,
            server=self._account.server_url
        )
        header = self._account.request_header
        # the accept field of the header needs to be updated not the usual JSON
        header['Content-Type'] = 'application/x-www-form-urlencoded'
        return self._account.send_request(url,
                                          headers=header,
                                          data=msg.as_server_request,
                                          post=True,
                                          expected_response=204)

    def trigger_remote_vehicle_finder(self) -> RemoteServiceStatus:
        """Trigger the vehicle finder.

        A state update is triggered after this, as the location state of the vehicle changes.
        """
        _LOGGER.debug('Triggering remote vehicle finder')
        # needs to be called via POST, GET is not working
        self._trigger_remote_service(_Services.REMOTE_VEHICLE_FINDER, post=True)
        result = self._block_until_done(_Services.REMOTE_VEHICLE_FINDER)
        self._trigger_state_update()
        return result



"""Test for ChargingProfile"""
import unittest
from unittest import mock
from test import load_response_json
from bimmer_connected.account import ConnectedDriveAccount
from bimmer_connected.state import VehicleState
from bimmer_connected.const import SERVICE_CHARGING_PROFILE
from bimmer_connected.charging_profile import ChargingMode, ChargingPreferences, TimerTypes

I01_TEST_DATA = load_response_json('I01_NOREX/charging_profile.json')


class TestState(unittest.TestCase):
    """Test for ChargingProfile."""

    # pylint: disable=protected-access

    def test_parse_i01(self):
        """Test if the parsing of the attributes is working."""
        account = mock.MagicMock(ConnectedDriveAccount)
        state = VehicleState(account, None)
        state._attributes[SERVICE_CHARGING_PROFILE] = I01_TEST_DATA['weeklyPlanner']
        self.assertTrue(state.charging_profile.is_pre_entry_climatization_enabled)
        self.assertEqual(ChargingMode.DELAYED_CHARGING, state.charging_profile.charging_mode)
        self.assertEqual(ChargingPreferences.CHARGING_WINDOW, state.charging_profile.charging_preferences)

        self.assertTrue(state.charging_profile.pre_entry_climatization_timer[TimerTypes.TIMER_1].timer_enabled)
        self.assertEqual('07:30',
                         state.charging_profile.pre_entry_climatization_timer[TimerTypes.TIMER_1].departure_time)
        self.assertEqual('MONDAY', state.charging_profile.pre_entry_climatization_timer[TimerTypes.TIMER_1].weekdays[0])

        self.assertEqual('05:02', state.charging_profile.prefered_charging_window.start_time)
        self.assertEqual('17:31', state.charging_profile.prefered_charging_window.end_time)



"""Models the navigation of a vehicle."""

import logging

from bimmer_connected.const import SERVICE_NAVIGATION

_LOGGER = logging.getLogger(__name__)


def backend_parameter(func):
    """Decorator for parameters reading data from the backend.

    Errors are handled in a default way.
    """
    def _func_wrapper(self: 'Navigation', *args, **kwargs):
        # pylint: disable=protected-access
        if self._state.attributes[SERVICE_NAVIGATION] is None:
            raise ValueError('No data available for vehicles navigation!')
        try:
            return func(self, *args, **kwargs)
        except KeyError:
            _LOGGER.debug('No data available for attribute %s!', str(func))
            return None
    return _func_wrapper


class Navigation:  # pylint: disable=too-many-public-methods
    """Models the navigation of a vehicle."""

    def __init__(self, state):
        """Constructor."""
        self._state = state

    @property
    @backend_parameter
    def attributes(self) -> dict:
        """Retrieve all attributes from the sever.

        This does not parse the results in any way.
        """
        return self._state.attributes[SERVICE_NAVIGATION]

    def __getattr__(self, item):
        """Generic get function for all backend attributes."""
        return self._state.attributes[SERVICE_NAVIGATION][item]

    @property
    @backend_parameter
    def latitude(self) -> float:
        """Returns the latitude."""
        return float(self._state.attributes[SERVICE_NAVIGATION]['latitude'])

    @property
    @backend_parameter
    def longitude(self) -> float:
        """Returns the longitude."""
        return float(self._state.attributes[SERVICE_NAVIGATION]['longitude'])

    @property
    @backend_parameter
    def iso_country_code(self) -> str:
        """Returns the iso country code."""
        return self._state.attributes[SERVICE_NAVIGATION]['isoCountryCode']

    @property
    @backend_parameter
    def aux_power_regular(self) -> float:
        """Returns the aux power regular consumption."""
        return float(self._state.attributes[SERVICE_NAVIGATION]['auxPowerRegular'])

    @property
    @backend_parameter
    def aux_power_eco_pro(self) -> float:
        """Returns the aux power eco pro consumption."""
        return float(self._state.attributes[SERVICE_NAVIGATION]['auxPowerEcoPro'])

    @property
    @backend_parameter
    def aux_power_eco_pro_plus(self) -> float:
        """Returns the aux power eco pro plus consumption."""
        return float(self._state.attributes[SERVICE_NAVIGATION]['auxPowerEcoProPlus'])

    @property
    @backend_parameter
    def soc(self) -> float:
        """Returns the soc."""
        return float(self._state.attributes[SERVICE_NAVIGATION]['soc'])

    @property
    @backend_parameter
    def soc_max(self) -> float:
        """Returns the soc max."""
        try:
            return float(self._state.attributes[SERVICE_NAVIGATION]['socMax'])
        except KeyError:
            return float(self._state.attributes[SERVICE_NAVIGATION]['socmax'])

    @property
    @backend_parameter
    def eco(self) -> str:
        """Returns the eco."""
        return self._state.attributes[SERVICE_NAVIGATION]['eco']

    @property
    @backend_parameter
    def norm(self) -> str:
        """Returns the norm."""
        return self._state.attributes[SERVICE_NAVIGATION]['norm']

    @property
    @backend_parameter
    def eco_ev(self) -> str:
        """Returns the eco ev."""
        return self._state.attributes[SERVICE_NAVIGATION]['ecoEv']

    @property
    @backend_parameter
    def norm_ev(self) -> str:
        """Returns the norm ev."""
        return self._state.attributes[SERVICE_NAVIGATION]['normEv']

    @property
    @backend_parameter
    def vehicle_mass(self) -> int:
        """Returns the vehicle mass."""
        return int(self._state.attributes[SERVICE_NAVIGATION]['vehicleMass'])

    @property
    @backend_parameter
    def k_acc_reg(self) -> int:
        """Returns the k acc reg."""
        return int(self._state.attributes[SERVICE_NAVIGATION]['kAccReg'])

    @property
    @backend_parameter
    def k_dec_reg(self) -> int:
        """Returns the k dec reg."""
        return int(self._state.attributes[SERVICE_NAVIGATION]['kDecReg'])

    @property
    @backend_parameter
    def k_acc_eco(self) -> int:
        """Returns the k acc eco."""
        return int(self._state.attributes[SERVICE_NAVIGATION]['kAccEco'])

    @property
    @backend_parameter
    def k_dec_eco(self) -> int:
        """Returns the k dec eco."""
        return int(self._state.attributes[SERVICE_NAVIGATION]['kDecEco'])

    @property
    @backend_parameter
    def k_up(self) -> int:
        """Returns the k up."""
        return int(self._state.attributes[SERVICE_NAVIGATION]['kUp'])

    @property
    @backend_parameter
    def k_down(self) -> int:
        """Returns the k down."""
        return int(self._state.attributes[SERVICE_NAVIGATION]['kDown'])

    @property
    @backend_parameter
    def drive_train(self) -> str:
        """Returns the drive_train."""
        return self._state.attributes[SERVICE_NAVIGATION]['driveTrain']

    @property
    @backend_parameter
    def pending_update(self) -> bool:
        """Returns the pending update state."""
        return bool(self._state.attributes[SERVICE_NAVIGATION]['pendingUpdate'])

    @property
    @backend_parameter
    def vehicle_tracking(self) -> bool:
        """Returns the vehicle tracking state."""
        return bool(self._state.attributes[SERVICE_NAVIGATION]['vehicleTracking'])



"""Get the right url for the different countries."""
from enum import Enum
import logging
from typing import List

_LOGGER = logging.getLogger(__name__)


class Regions(Enum):
    """Regions of the world with separate servers."""
    NORTH_AMERICA = 0
    CHINA = 1
    REST_OF_WORLD = 2


#: Mapping from regions to servers
_SERVER_URLS = {
    Regions.NORTH_AMERICA: 'b2vapi.bmwgroup.us',
    Regions.REST_OF_WORLD: 'b2vapi.bmwgroup.com',
    Regions.CHINA: 'b2vapi.bmwgroup.cn:8592'
}

#: Mapping from regions to servers
_GCDM_OAUTH_ENDPOINTS = {
    Regions.NORTH_AMERICA: 'customer.bmwgroup.com/gcdm/usa',
    Regions.REST_OF_WORLD: 'customer.bmwgroup.com/gcdm',
    Regions.CHINA: 'customer.bmwgroup.cn/gcdm'
}


def valid_regions() -> List[str]:
    """Get list of valid regions as strings."""
    return [k.lower() for k in Regions.__members__.keys()]


def get_region_from_name(name: str) -> Regions:
    """Get a region for a string.

    This function is not case-sensitive.
    """
    for region_name, region in Regions.__members__.items():
        if name.lower() == region_name.lower():
            return region
    raise ValueError('Unknown region {}. Valid regions are: {}'.format(
        name,
        ','.join(valid_regions())))


def get_server_url(region: Regions) -> str:
    """Get the url of the server for the region."""
    return _SERVER_URLS[region]


def get_gcdm_oauth_endpoint(region: Regions) -> str:
    """Get the url of the server for the region."""
    return _GCDM_OAUTH_ENDPOINTS[region]
